
/**
 * Annotation.java
 *
 * This class implements an annotation, which is a text comment
 * associated with a physical area on a bitmap - represented by
 * a shape of some kind.  Right now those shapes are GeneralPath
 * objects, but we would also like to be able to have ovals and
 * rectangles.  These will be implemented in the near future.  
 *
 * The implementation of the shape this Annotation covers is done
 * using the java.awt.geom.shape abstract class.  We then instantiate
 * one of "GeneralPath" (the first implementation this worked with), a
 * "Rectangle" or a "Ellipse2D.Double".  The drawSelf method then
 * draws the appropriate shape based on the instantiated shape object.
 *
 * A specific annotator is also associated with each annotation.
 *
 * Now we have the joy of having to be able to supply scaled versions
 * of the Annotations as well.  We do that by keeping around a 
 * scaledShape reference.  If it is null, we don't have a scaled 
 * version.  If it is not null, we assume that it is the scale that
 * we are currently at.  When our scale(int scale) method is called,
 * we will either create the new scaledShape based on scale and our
 * shape, or we will set scaledShape to null if scale is 100.
 *
 * LOTD:
 *   make a toString
 *
 * Created: Thu Mar 09 00:24:44 2000
 *
 * @author David Evans
 * @version 0.1 alpha */


/* 
   GMAO - Image Annotation for Manga Translation
   Copyright (C) 2000  David Kirk Evans
   
   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

   David Kirk Evans (devans@cs.columbia.edu)
   http://www.cs.columbia.edu/~devans/   
*/


/* Import packages */

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import java.awt.geom.*;
import java.awt.image.BufferedImage;
import java.util.*;

/* Import other GMAO2k classes */
import Annotator;

public class Annotation implements java.io.Serializable {
    Vector notes;  // The notes
    Shape location;  // The shape that this annotation covers. 
    boolean fill = false;  // Should we fill things in or not?
    Color fillColor = Color.white;  // Color to fill in with.
    Shape scaledShape=null;  // Scaled version of shape.
    /**
     *  Create an empty annotation that is just an area.  Notes 
     *  can be attached to this area later, where each note tracks
     *  a string that is the annotation, and a Annotator for the
     *  person who made the annotation.  
     */
    public Annotation(Shape shape) {
	location = shape;
	notes = new Vector();
    }
    
    /**
     *  drawSelf will draw the shape of the annotation onto the 
     *  passed in graphics2d.  It will all the types of self that
     *  it knows how.  (So no matter what shape it is, it will draw it.)
     *  Have to check this out, but I think that since we are using 
     *  shape and Graphics2D, we don't need to change the code at all
     *  for the different shape cases.  
     */
    public void drawSelf(Graphics2D g) {
	if (location instanceof GeneralPath) {
	    if (fill) {
		Color tempColor = g.getColor();
		g.setColor(fillColor);
		if (scaledShape == null) {
		    g.fill((GeneralPath) location);
		} else {
		    g.fill((GeneralPath) scaledShape);
		}
		// return color to normal
		g.setColor(tempColor);
	    }
	    if (scaledShape == null) {
		g.draw((GeneralPath) location);
	    } else {
		g.draw((GeneralPath) scaledShape);
	    }	
	} else if (location instanceof Rectangle2D.Double) {
	    if (fill) {
		Color tempColor = g.getColor();
		g.setColor(fillColor);
		if (scaledShape == null) {
		    g.fill((Rectangle2D.Double) location);
		} else {
		    g.fill((Rectangle2D.Double) scaledShape);
		}
		// return color to normal
		g.setColor(tempColor);
	    }
	    if (scaledShape == null) {
		g.draw((Rectangle2D.Double) location);	    
	    } else {
		g.draw((Rectangle2D.Double) scaledShape);
	    }
	} else if (location instanceof Ellipse2D.Double) {
	    if (fill) {
		Color tempColor = g.getColor();
		g.setColor(fillColor);
		if (scaledShape == null) {
		    g.fill((Ellipse2D.Double) location);
		} else {
		    g.fill((Ellipse2D.Double) scaledShape);
		}
		// return color to normal
		g.setColor(tempColor);
	    }
	    if (scaledShape == null) {		
		g.draw((Ellipse2D.Double) location);	    
	    } else {
		g.draw((Ellipse2D.Double) scaledShape);
	    }
	}
	// implement other types of supported shapes here.
    }

    /**
     *  drawSelf will draw the shape of the annotation onto the 
     *  passed in graphics2d.  It will all the types of self that
     *  it knows how.  (So no matter what shape it is, it will draw it.)
     *  Have to check this out, but I think that since we are using 
     *  shape and Graphics2D, we don't need to change the code at all
     *  for the different shape cases.  
     *
     *  This one also sets the fill color for this shape
     *
     *  And will draw the text text, or not if text is null
     *  This is really inefficient right now.  I'm re-calculating 
     *  where the strings should go every time it is re-painted.
     *  I should probably cache these once they have been calculated
     *  once.  Anyway.  
     */
    public void drawSelf(Graphics2D g, boolean fillValue,Color fillColorValue, String text) {
	setFillColor(fillColorValue);
	setFill(fillValue);
	drawSelf(g);
	if (text != null) {
	    // Get the bounding box for this Shape
	    Shape myShape = getShape();
	    if (myShape instanceof GeneralPath) {
		// Punt here.  
		return;
	    }
	    Rectangle2D.Double bounds = (Rectangle2D.Double)myShape.getBounds2D();
	    // Use a string tokenizer to break the text up into chunks
	    // We want to return the tokens so we can put the spaces back in...
	    StringTokenizer tok = new StringTokenizer(text, "\t\n\r\f ", true);
	    // Use an ArrayList to keep a few StringBuffers around
	    ArrayList bufs = new ArrayList(10);
	    // Points for where to draw each string at
	    ArrayList coords = new ArrayList(10);
	    StringBuffer buf = new StringBuffer();
	    // Point for the "center-top" of this object 
	    double cx = bounds.x + (bounds.width / 2);
	    double cy = bounds.y+1;
	    Point2D.Double center = new Point2D.Double(cx, cy);
	    FontMetrics fm = g.getFontMetrics();
	    // Try to put tokens into the buffer one "line" per index
	    // Try to fill lines down from the top.  	    
	    int i=0;
	    // Height of a standard line.  
	    int lineHeight = fm.getHeight();
	    while (tok.hasMoreTokens()) {
		String token = tok.nextToken();
		// Test to see if the token can go on current line
		if (i == bufs.size()) {
		    bufs.add(i, new StringBuffer(""));
		}
		StringBuffer sb = (StringBuffer)bufs.get(i);
		if (i == coords.size()) {
		    // Find out where to put it - start at center, move left
		    // in increments of 2 pixels say, put on right line
		    double py = cy + lineHeight*(i);
		    double px = cx;  		    
		    // Check for the case that we are short, and lineHeight
		    // is > height, knocking us out just because of vertical
		    if (lineHeight > bounds.getHeight()) {
			py = bounds.y + bounds.getHeight();  // Set us to bottom of box
		    }
		    while (myShape.contains(px, py, 2, lineHeight)) {
			px = px - 2;
		    }
		    coords.add(i, new Point2D.Double(px+2, py));
		}
		Rectangle2D textBounds = fm.getStringBounds(sb.toString() + token, g);
		// Check to see if the textBounds box at point fits in
		textBounds.setRect(((Point2D.Double)coords.get(i)).x, ((Point2D.Double)coords.get(i)).y, textBounds.getWidth(), textBounds.getHeight());
		Graphics2D g2 = (Graphics2D) g;
		if (myShape.contains(textBounds)) {
		    // Token can go on current line
		    sb.append(token);	
		    /* Color temp = g2.getColor();
		     * g2.setColor(Color.green);
		     * g2.draw(textBounds);
		     * g2.setColor(temp);
		     */
		} else {
		    // Token has to go on another line. 
		    /* Color temp = g2.getColor();
		     * g2.setColor(Color.red);
		     * g2.draw(textBounds);
		     * g2.setColor(temp);
		     * System.out.println(textBounds + " didn't fit");
		     */
		    i++;
		    // Put token in
		    bufs.add(i, new StringBuffer(token));
		    // Find out where to put it - start at center, move left
		    // in increments of 2 pixels say, put on right line
		    double py = cy + lineHeight*(i);
		    double px = cx;  
		    // Check for the case that we are short, and lineHeight
		    // is > height, knocking us out just because of vertical
		    if (lineHeight > bounds.getHeight()) {
			py = bounds.y + bounds.getHeight();  // Set us to bottom of box
		    }
		    while (myShape.contains(px, py, 2, lineHeight)) {
			px = px - 2;
		    }
		    coords.add(i, new Point2D.Double(px+2, py));
		}
	    }
	    // System.out.println("Myshape: "+myShape);
	    // System.out.println("Drawing: "+bufs);
	    // System.out.println("At: "+coords);
	    // Now draw the resultsing strings
	    for (int j=0; j<bufs.size(); j++) {
		Point2D.Double coord = (Point2D.Double) coords.get(j);
		int xi = (int)coord.x;
		// Have to specify the bottom line where we want to draw
		// the point's y is the upper limit - so add in lineHeight
		int yi = (int)coord.y+lineHeight;
		g.drawString(((StringBuffer)bufs.get(j)).toString(), xi, yi);
	    }
	}
    }

    public boolean getFill() {
	return fill;
    }

    public void setFill(boolean b) {
	fill = b;
    }

    public void setFillColor(Color f) {
	fillColor = f;
    }

    public Color getFillColor() {
	return fillColor;
    }

    public Shape getShape() {
	if (scaledShape == null)
	    return location;
	else 
	    return scaledShape;
    }

    public Note getNote(int num) {
	return (Note) notes.elementAt(num);
    }

    /**
     *  Note getNote(String annotator, String language) 
     *
     *  This method will search for the notes attached to this 
     *  annotation and return the one that best matches the above
     *  parameters.  If there is a note with both the specified
     *  annotator and the specified language, it is returned. 
     *  Otherwise, language takes precedence, and the last note
     *  of the correct language will be returned.  If there is no
     *  note of that language, the last note by the annotator is
     *  returned.  If there are no notes by that annotator, the first
     *  note is returned.
     */
    public Note getNote(String annotator, String language) {
	Note theNote = null;
	int num=0;
	// Search for the note with the right language and annotator.
	while (num < notes.size()) {
	    if (language.equals(((Note)notes.elementAt(num)).getCustomFieldValue("Language"))) {
		// We have the rigth language.  
		theNote = (Note)notes.elementAt(num);
		// Check to see if the annotator is right
		if (theNote.getAnnotator().getID().equals(annotator)) {
		    // It matches.  Return this note.
		    return theNote;
		}
		// Not the right annotator.  Keep searching.  
	    }
	    // Check the next one
	    num++;
	}
	// Well, if theNote isn't null it is either the last 
	// matching language, or the last matching annotator.
	// Return it, unless it is null
	if (theNote == null && notes.size() > 0) {
	    // Return the first note we have
	    return (Note) notes.elementAt(0);
	} else if (notes.size() == 0) {
	    return null;
	} else {
	    return theNote;
	}
    }

    public void setNote(int num, Note s) {
	//  Check to make sure that we have a note number num to set!
	if (num >= 0 && num < notes.size()) {
	    notes.setElementAt(s, num);
	} else if (num >=0 && num == notes.size()) {	    
	    //  We need to make a new note
	    notes.add(s);
	} else {
	    // Debugging stuff
	    System.out.println("Tried to set note number "+num+" when we only have " + notes.size() + " notes.");
	}
    }    

    public void addNote(Note n) {
	notes.add(n);
    }

    public int getNumNotes() {
	return notes.size();
    }        

    public void scale(int scale) {
	if (scale == 100) {
	    // Remove scaled version of self
	    scaledShape = null;
	} else {
	    // Must create a new scaledShape to scale of scale
	    if (location instanceof GeneralPath) {
		scaledShape = ((GeneralPath)location).createTransformedShape(AffineTransform.getScaleInstance(scale / 100, scale / 100));
	    } else if (location instanceof Rectangle2D.Double) {
		double scaleD = ((double)scale) / ((double)100);
		scaledShape = new Rectangle2D.Double(((Rectangle2D.Double)location).getX() * scaleD, ((Rectangle2D.Double)location).getY() * scaleD, ((Rectangle2D.Double)location).getWidth() * scaleD, ((Rectangle2D.Double)location).getHeight() * scaleD);
	    } else if (location instanceof Ellipse2D.Double) {
		double scaleD = ((double)scale) / ((double)100);
		scaledShape = new Ellipse2D.Double(((Ellipse2D.Double)location).getX() * scaleD, ((Ellipse2D.Double)location).getY() * scaleD, ((Ellipse2D.Double)location).getWidth() * scaleD, ((Ellipse2D.Double)location).getHeight() * scaleD);
	    }
	}
    }
} // Annotation









