
/**
 * AnnotationInformationPanel.java
 *
 * A Panel that displays information about an annotation, and allows
 * for editing and creation of new notes for an annotation.  
 *
 * A current Annotation is set as the one to display.  We get a JPanel
 * from the individual Notes in the Annotation, and display that.  We
 * also have to control which note to display (using prev and next
 * buttons when applicable.)
 *
 * When creating a new Note, we throw up a Modal Dialog that prompts
 * for the information for the new Note.  We are responsible for
 * creating that modal dialog though, and not the Note object.
 *
 * Similarly, when we Edit an existing Note, we through up a modal
 * dialog and prompt for information.  Actually, this is exactly the
 * same as the case in the creation of a new Note, but we place the
 * existing values into the fields, instead of leaving them blank.  So
 * it shoudl be essentially the same function call, with maybe a
 * prompt string here or there different.
 *
 * Created: Thu Mar 09 09:51:03 2000
 *
 * @author David Evans
 * @version 0.2 alpha */

/* 
   GMAO - Image Annotation for Manga Translation
   Copyright (C) 2000  David Kirk Evans
   
   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

   David Kirk Evans (devans@cs.columbia.edu)
   http://www.cs.columbia.edu/~devans/   
*/


/* Import packages */

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import java.awt.geom.*;
import java.awt.image.BufferedImage;
import java.util.*;
import java.net.URL;
import java.text.DateFormat;

/* Import other GMAO2k stuff */
import Annotator;
import Annotation;
import Note;

public class AnnotationInformationPanel extends JPanel implements ActionListener {
    Annotation current;  // Annotation to display
    boolean editable = false;  // by default, just display, don't edit.  
    JLabel noteNum;  // The number of the note for this annotation
    int currentNoteNumber = 0;  //  The current note number we are displaying.
    JButton next;  // Button on the main display
    JButton prev;  // Button on the main display
    JButton edit;  // Button on the main display
    JButton newButton;  // Button on the main display
    /**
     * JPanel centerDisplayed is just a pointer to the current JPanel
     * in the center of the frame.  We need this because whenver we
     * display a new note, we remove the JPanel from the center that
     * we are showing - using this pointer.  When we create a new
     * panel for the new Note, we set centerDisplayed to it. */
    JPanel centerDisplayed; 
    /**
     * JEditorPane display is the display where the information goes.
     * It is set up to work with HTML content.  The notes are 
     * then asked to return a HTML string that represents their content.
     */
    JEditorPane display;  
    /**
     * JFrame ownerFrame is kept around solely because we need to know
     * what frame we are in to create the Dialog to edit/create new
     * notes with.  */
    JFrame ownerFrame; 
    /**
     * JDialog editDialog is the JDialog that we put up for editing or
     * creating new Notes.  We need to have a ref to it around so that
     * we can close it when appropriate.  There might be problems
     * right now with closing the dialog with the close button on the
     * title bar as opposed to the "ok" or "cancel" buttons on the
     * dialog - I have not checked into that - nor added a window
     * listener to the thing.  */
    JDialog editDialog;
    /**
     *  Note dialogNote is kept around as the result of the last
     *  action in the create/edit dialog box.  If the user cancelled
     *  their action, this will be null, a value that we can check on
     *  later.  If the user filled in all the fields, then this will
     *  be a new Note that reflects the values in the fields.  We also
     *  can use that later.  */
    Note dialogNote = null;  // The note that is modified by editNoteDialog
    /**
     *  JTextArea dialogNoteTextArea is the text area that we enter
     *  the actual note into when we are editing an existing note or
     *  creating a new on in the editNoteDialog.  We create this once
     *  in the constructor, and from then on just set the text to
     *  different values.  When I get some time I'll add a border
     *  around this puppy to make it look nice.  */
    JTextArea dialogNoteTextArea;  
    /**
     *  Vector dialogCustomFieldTextFields is a vector that consists of
     *  JTextFields that corresspond to the entries in Annotation
     *  current.getNote(currentNoteNumber).getCustomFields() --
     *  essentially the set of custom fields for the current note of
     *  the current annotation.  */
    Vector dialogCustomFieldTextFields;
    String[] customFields = {"Language", "Character"};
    String[] customDefaults = {"English", ""};
    Annotator currentAnnotator;
    AnnotatorPanel annPanel = null;

    public AnnotationInformationPanel(JFrame owner) {
	/*  Use a border layout.  We have these UI elements:
	 *  Buttons: Note: #, Previous, Next, Edit, New
	 *  and a big panel in the center for the Note display
	 */
	dialogNoteTextArea = new JTextArea(20, 30);
	//  Make it wrap words
	dialogNoteTextArea.setLineWrap(true);
	dialogNoteTextArea.setWrapStyleWord(true);
	ownerFrame = owner;  
	//  Add the buttons to the top of a border layout	
	setLayout(new BorderLayout());
	JPanel northPanel = new JPanel();
	//  Put the number of the note displayed and total number of
	//  notes here
	JLabel noteNumLabel = new JLabel("Note #:");
	noteNum = new JLabel("none");
	northPanel.add(noteNumLabel);
	northPanel.add(noteNum);
	//  Add buttons to get next and previous notes
	prev = new JButton("Previous");
	prev.addActionListener(this);
	northPanel.add(prev);
	next = new JButton("Next");
	next.addActionListener(this);
	northPanel.add(next);
	// South panel
	JPanel southPanel = new JPanel();
	edit = new JButton("Edit");
	edit.setMnemonic('e');
	edit.addActionListener(this);
	southPanel.add(edit);
	newButton = new JButton("New");
	newButton.setMnemonic('n');
	newButton.addActionListener(this);
	southPanel.add(newButton);
	add(northPanel, BorderLayout.NORTH);
	add(southPanel, BorderLayout.SOUTH);
	// Set up the display in the CENTER
	display = new JEditorPane("text/html", "");
	display.setEditable(false);
	add(display, BorderLayout.CENTER);
	updateDisplay();
    }

    public AnnotationInformationPanel(Annotation c, boolean edit, JFrame owner, Annotator ca) {
	this(owner);  // call lower constructor
	current = c;
	editable = edit;
	currentAnnotator = ca;
	updateDisplay();	
    }
    
    public void setAnnotation(Annotation c) {
	current = c;
	currentNoteNumber = 0;
	updateDisplay();
    }

    public Annotation getAnnotation() {
	return current;
    }

    public void updateDisplay() {	
	//System.out.println("Updating Display...");
	if (current != null && current.getNumNotes() != 0) {
	    //  Set Note Number label
	    noteNum.setText(""+(currentNoteNumber+1)+"/"+current.getNumNotes());
	    //  Remove the currently displayed panel
	    /*  Removed this code: switching to using a JEditorPane
	    if (centerDisplayed != null) {
		remove(centerDisplayed);
		centerDisplayed = null;
	    }
	    //  Display the display panel for the current note here
	    centerDisplayed = ((Note)current.getNote(currentNoteNumber)).getDisplayPanel();
	    add(centerDisplayed, BorderLayout.CENTER);
	    */
	    
	    // Change the text in display
	    display.setText(((Note)current.getNote(currentNoteNumber)).getHTMLRepresentation());
	    validate();
	} else {
	    //  There is no annotation 
	    noteNum.setText("none");
	    //  Remove text from the display
	    display.setText("");
	}
	//  Update button display
	updateButtonDisplay();
	invalidate();
	repaint();
    }
    
    /** 
     *  Make sure that the Prev and Next buttons are only 
     *  enabled when they are supposed to be.  
     */
    public void updateButtonDisplay() {
	if (current == null || currentNoteNumber == 0) {
	    // No notes or no annotation - disable prev
	    prev.setEnabled(false);
	} else {
	    prev.setEnabled(true);
	}
	if (current == null || current.getNumNotes() == 0 || currentNoteNumber == (current.getNumNotes() - 1)) {
	    //  At last note, or no note to display or no annotation at all - disable next
	    next.setEnabled(false);
	} else {
	    next.setEnabled(true);
	}
	//  if we have no annotation don't let us make a new one!
	if (current == null) {
	    newButton.setEnabled(false);
	} else {
	    newButton.setEnabled(true);
	}
	//  If we have an annotation but no note don't let us edit!
	if (current == null || current.getNumNotes() == 0) {
	    edit.setEnabled(false);
	} else {
	    edit.setEnabled(true);
	}
    }

    public void actionPerformed(ActionEvent e) {
	/* Look for actions: Next, Previous, Edit, New
	 */
	String action = e.getActionCommand();
	if (action.equals("Next")) {
	    //  Put up the next note
	    currentNoteNumber++;
	    updateDisplay();
	} else if (action.equals("Previous")) {
	    //  Put up the previous note
	    currentNoteNumber--;
	    updateDisplay();
	} else if (action.equals("Edit")) {
	    if (currentAnnotator == null) {
		JOptionPane.showMessageDialog(this, "You must log in as an annotator to edit translations", "Please log in", JOptionPane.ERROR_MESSAGE); 
		return;
	    }
	    //  Update the information in the note
	    if (current != null) {
		// Check to see if the annotator for the note is the currentAnnotator
		if (!current.getNote(currentNoteNumber).getAnnotator().getID().equals(currentAnnotator.getID())) {
		    JOptionPane.showMessageDialog(this, "You may only change annotations that you created.  This change will not be saved on the remote server.", "Changes will not be saved", JOptionPane.WARNING_MESSAGE); 
		}
		//  Have to make an Edit Note Dialog box here
		Note newNote = editNoteDialog("Edit Note #"+currentNoteNumber+1, current.getNote(currentNoteNumber), currentNoteNumber+1);
		//  Add the edited note - returns null on Cancel
		if (newNote != null) {
		    current.setNote(currentNoteNumber, newNote);
		    // Alert the annotation panel that this note has changed
		    annPanel.annotationUpdated(current, currentNoteNumber);
		}
	    } else {
		JOptionPane.showMessageDialog(null, "You must select an Annotation!", "alert", JOptionPane.ERROR_MESSAGE); 
	    }
	    updateDisplay();
	} else if (action.equals("New")) {
	    //  Check to make sure we have an annotator
	    if (currentAnnotator == null) {
		JOptionPane.showMessageDialog(this, "You must log in as an annotator to create new translations", "Please log in", JOptionPane.ERROR_MESSAGE); 
		return;
	    }
	    //  Create a new note	    
	    if (current != null) {
		//  Increment the note number - unless there are no notes
		if (current.getNumNotes() > 0) 
		    currentNoteNumber++;
		else 
		    currentNoteNumber = 0;
		//  Have to make a New Note Dialog box here
		Note newNote = editNoteDialog("Create a new Note", null, currentNoteNumber+1);
		//  Add the new note in to current - returns null on cancel
		if (newNote != null) {
		    current.setNote(currentNoteNumber, newNote);
		    /* Also, here we have to add the new note to 
		       the changed annotations in the AnnotatePanel but
		       only if it is not already in the changes vector...
		    */
		    annPanel.annotationUpdated(current, currentNoteNumber);
		} else {
		    //  Have to remember that we really didn't add it...
		    currentNoteNumber--;
		}
	    } else {
		JOptionPane.showMessageDialog(null, "You must select an Annotation!", "alert", JOptionPane.ERROR_MESSAGE); 
	    }
	    updateDisplay();
	} else if (action.equals("OK")) {
	    //  OK From the Dialog for a New or Edit 
	    //  We have to create a new note based on the fields in 
	    //  in the DialogBox.  	    
	    //  Make a new vector for the new custom fields
	    boolean done = true;  //  Only be done if all fields filled out.
	    Vector newCustFields = new Vector();
	    String newNoteText = dialogNoteTextArea.getText();
	    //  Can't be done if they haven't put in the note text!
	    if (newNoteText.equals("")) {
		done = false; 
	    }
	    //  For each custom field add it in
	    for (int i=0; i<dialogCustomFieldTextFields.size(); i++) {
		//  Here we have to check that each custom field that must
		//  be filled out, is filled out.  If not, set done to false.
		//  Annotation has to be changed to have the customfields too. 
		if (((JTextField)dialogCustomFieldTextFields.elementAt(i)).getText().equals("")) {
		    // Should really check to see if it is required.  But we can't really...
		    done = false;
		} else {
		    // Add the field to the newCustFields
		    newCustFields.add(new NoteCustomField(customFields[i], ((JTextField)dialogCustomFieldTextFields.elementAt(i)).getText(), true));
		    // If it is the "Character" field, and it is not in 
		    // TextCompleter, add it in.
		    if (customFields[i].equals("Character")) {
			// Just add it in anyway. 
			TextCompleter.addText(((JTextField)dialogCustomFieldTextFields.elementAt(i)).getText());
		    }
		}
	    }
	    if (done) {
		dialogNote = new Note(newNoteText, currentNoteNumber+1, newCustFields);
		//  Close the window
		editDialog.dispose();
	    } else {
		//  Future feature: tell them what fields are missing!
		JOptionPane.showMessageDialog(null, "You must fill out all required fields!", "alert", JOptionPane.ERROR_MESSAGE); 
		dialogNote = null;
	    }
	} else if (action.equals("Cancel")) {
	    //  Cancel from the Dialog for a New or Edit
	    //  All we have to do is set dialogNote to null.  
	    dialogNote = null;
	    //  Close the window
	    editDialog.dispose();
	}
    }
    
    /**
     *
     *  create a modal dialog with title title, that allows us to
     *  create a new Note based on the values already in Note theNote,
     *  and this note will be number noteNum.  Be sure to require all
     *  required fields, as specified in the current Annotation's
     *  customFields vector.  Return "null" on a Cancel action,
     *  otherwise, return the new Note.
     *
     *  Also, put the editable note part into a scroll pane.  
     */
    Note editNoteDialog(String title, Note theNote, int noteNum) {
	//  Create the dialog that the frame we are in owns, with the
	//  passed in title, and that is modal
	editDialog = new JDialog(ownerFrame, title, true);
	if (theNote == null) {
	    theNote = new Note("", noteNum);
	    //  Ok, create CUSTOM FIELDS HERE
	    for (int i=0; i<customFields.length; i++) {
		theNote.addCustomField(new NoteCustomField(customFields[i], customDefaults[i], true));	    
	    }
	}
	//  At any rate, we want to display the note number and the
	//  text of the note
	//  Set the layout Manager for me to be a borderlayout
	editDialog.getContentPane().setLayout(new BorderLayout());
	//  Add to the top a BoxLayout with Note and Date
	Box temp = new Box(BoxLayout.X_AXIS);
	temp.add(new JLabel("Note #"+theNote.getNumber()));
	//  Put in some Horizontal glue there.  
	temp.add(Box.createHorizontalGlue());
	temp.add(new JLabel("Created: " + DateFormat.getDateInstance().format(theNote.getDate())));
	//  Add the number/date panel to me in the NORTH
	editDialog.getContentPane().add(temp, BorderLayout.NORTH);
	//  Now the harder part: The center should be any custome fields
	//  that are defined, taking up minimum vertical space.  Split
	//  that with a big text area for the actual annotation.  	
	Box smallBox = new Box(BoxLayout.Y_AXIS);
	//  Check to see if we have any customfields

	Vector customFields = theNote.getCustomFields();
	//  Make a new Vector to hold the new custom field text fields
	dialogCustomFieldTextFields = new Vector();
	for (int i=0; i<customFields.size(); i++) {
	    //  Add in custom field at i
	    NoteCustomField field = (NoteCustomField) customFields.elementAt(i);
	    temp = new Box(BoxLayout.X_AXIS);
	    temp.add(new JLabel(field.getName()+":"));
	    JTextField tempField = new JTextField(field.getValue());
	    //  If this is the Character field, let's make the document
	    //  for it a TextCompletionDocument
	    if (field.getName().equals("Character")) {
		tempField = new TextCompletionField();
		// How convoluted is this?  Tell tempField's document
		// that tempField wants to listen to it...
		tempField.getDocument().addDocumentListener((TextCompletionField)tempField);
		// Set up the field to have the initial value of the field...
		tempField.setText(field.getValue());
	    }
	    dialogCustomFieldTextFields.add(tempField);
	    temp.add(tempField);
	    //  Add it in
	    smallBox.add(temp);
	}
	//  Add in the text area for the note
	//  Put a small JLabel before this
	JLabel noteLabel = new JLabel("Note:");
	noteLabel.setAlignmentX(Component.CENTER_ALIGNMENT);
	smallBox.add(noteLabel);
	//  Set the text of the text area to the text from the note
	dialogNoteTextArea.setText(theNote.getNote());
	JPanel bigTemp = new JPanel();
	bigTemp.setLayout(new BorderLayout());
	bigTemp.add(smallBox, BorderLayout.NORTH);
	bigTemp.add(new JScrollPane(dialogNoteTextArea), BorderLayout.CENTER);
	//  Now add in the bigTemp to me
	editDialog.getContentPane().add(bigTemp);
	//  Add an OK and Cancel button to the SOUTH
	JButton OKButton = new JButton("OK");
	OKButton.addActionListener(this);
	JButton CancelButton = new JButton("Cancel");
	CancelButton.addActionListener(this);
	temp = new Box(BoxLayout.X_AXIS);
	//  Put in some glue
	temp.add(Box.createHorizontalGlue());
	temp.add(CancelButton);
	temp.add(OKButton);
	editDialog.getContentPane().add(temp, BorderLayout.SOUTH);
	//  Show the dialog.  Pack it first so that it has some 
	//  (albeit too small) size.  
	editDialog.pack();
	editDialog.show();

	//  Have to think about how this is going to work -
	//  the dialog will stick around until it is cancelled or 
	//  OKd.  So we should have a dialogNote in the class, and
	//  that will keep either a null or the new note depending on
	//  button outcome.  then we can just return that.  
	
	//  We might be null...  Check to see that we are not.
	if (dialogNote != null) {
	    dialogNote.setAnnotator(currentAnnotator);
	}

	return dialogNote;    
    }

    public void setNoteChangeListener(AnnotatorPanel ap) {
	annPanel = ap;
    }

    public void setAnnotator(Annotator a) {
	currentAnnotator = a;
    }

} // AnnotationInformationPanel
