
/**
 * GMAOCacheWatcher.java
 *
 *  This class is just a thread that checks the amount of free
 *  memory that we have, and when it gets "low", removes some
 *  images from the imgCache that it keeps a reference to.  
 *
 *  The images will be removed in alphabetical order.  So the images
 *  with lower lexicographic order are removed first.  (This should
 *  have the effect of removing images that are earlier in the view
 *  sequence, so the user should be less likely to want to go back to
 *  that page.)
 *
 * Created: Tue Jul 18 14:52:07 2000
 *
 * @author David Evans
 * @version 0.1B
 */

import java.util.*;
import javax.swing.*;
import java.lang.Runtime;
import java.lang.Thread;

public class GMAOCacheWatcher extends Thread {
    Hashtable imageCache;
    Properties props;
    /**
     * MEMORY_LIMIT tells us how much memory we should have free
     * If memory gets lower than this, we will remove images until
     * it is above this limit again.
     */
    int MEMORY_LIMIT = 1024;  
    int cacheSize=3;  // Number of images to keep in the cache
    public GMAOCacheWatcher(Hashtable imgCache, Properties p) {
	imageCache = imgCache;
	props = p;
	//  If cacheSize set in the properties file, take it
	if (props.getProperty("cacheSize") != null) {
	    try {
		cacheSize = Integer.parseInt((String)props.getProperty("cacheSize"));
		MEMORY_LIMIT = Integer.parseInt((String)props.getProperty("memoryLimit"));
	    } catch (Exception e) {
		System.out.println("Couldn't parse setting for cacheSize "+((String)props.getProperty("cacheSize"))+" or memoryLimit "+((String)props.getProperty("memoryLimit")));
	    }
	}
    }
    
    /**
     *  Changes the size of the cache
     */
    public void setCacheSize(int size) {
	cacheSize = size;
    }
    
    /**
     *  Removes a single image from the image cache that we are 
     *  watching.  Hope that is all we need to do to clear up the 
     *  memory problems.  Remove the lexicographically first image
     *  on the keys
     */
    public void removeAnImage() {
	Enumeration keys = imageCache.keys();
	String smallest = (String) keys.nextElement();
	String b;
	while (keys.hasMoreElements()) {
	    b = (String) keys.nextElement();
	    if (smallest.compareTo(b) > 0) {
		smallest = b;
	    }
	}
	// Smallest key is in smallest.  Remove it.
	ImageIcon img = (ImageIcon) imageCache.remove(smallest);
	// Set it to null to gc it
	img = null;
	// Put up a little note
	System.out.println("Removed image "+smallest+" from cache because we are low on memory.");
	AnnotatePanel.infoDisplay.setText("Removed image "+smallest+" from cache because we are low on memory.");
    }

    public void run() {
	System.out.println("Starting up GMAOCacheWatcher thread...");	
	Runtime runtime = Runtime.getRuntime();
	long freeMem, totalMem;
	while (true) {
	    freeMem = runtime.freeMemory();
	    totalMem = runtime.totalMemory();
	    // System.out.println("We have "+freeMem+" / "+totalMem+" bytes free.");
	    while (imageCache.size() > cacheSize) {
		// System.out.println("We only have "+freeMem+" mem available.  Purging cache...");
		// Clean up the imageCache
		if (imageCache.isEmpty()) {
		    // Have nothing to clean up
		    System.out.println("imageCache is empty, we are just low on memory.");
		    break;
		} else {
		    removeAnImage();
		}
		// Ask the system to free up some memory 
		System.gc();
		freeMem = runtime.freeMemory();
	    }
	    try {
		Thread.sleep(5000);
	    } catch (InterruptedException ie) {
		System.out.println("GMAOCacheWatcher interrupted from sleeping: "+ie);
	    }
	}
    }

} // GMAOCacheWatcher
