
/**
 * GMAOLanguagePairs.java
 *
 *  This is a class that contains information on which 
 *  langauge / annotator pairs exist (for a given DOM - 
 *  if used the way it is supposed to be used.)  
 *
 *  This class will track, for a given annotator which languages 
 *  that annotator has annotated, or for a given language, 
 *  which annotators have annotated that language.  Used for 
 *  setting the display options.  The user will have to select
 *  a language / annotator pair to display, and this will have
 *  them.  
 *
 *  We also track information in here about how many annotations
 *  have been made on a per language and per annotator basis.  
 *
 * Created: Thu Aug 17 20:17:42 2000
 *
 * @author David Evans
 * @version 0.1a
 */


/* 
   GMAO - Image Annotation for Manga Translation
   Copyright (C) 2000  David Kirk Evans
   
   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

   David Kirk Evans (devans@cs.columbia.edu)
   http://www.cs.columbia.edu/~devans/   
*/

import java.util.*;

public class GMAOLanguagePairs  {
    /** languageCounts is a hashtable that counts the number of
     *  annotations for each language.
     */
    Hashtable languageCounts;
    /** annotatorCounts is a hashtable that, given an annotatorID
     *  returns a hashtable that has the counts for the annoations 
     *  that annotator has made on a per-language basis. 
     */
    Hashtable annotatorCounts;
    
    public GMAOLanguagePairs() {
	languageCounts = new Hashtable();
	annotatorCounts = new Hashtable();	
    }
    
    /**
     *  This is the main data entry point for the class.  Given a 
     *  string that is the ID for some annotator, and a language that
     *  is the language the annotation is in, increment the associated
     *  hashtable counts. 
     *
     *  If any of annotatorID or language do not exist, they will be
     *  created. 
     */
    public void incrementAnnotationCount(String annotatorID, String language) {
	// Check to see if the language exists, update languageCounts
	if (languageCounts.get(language) == null) {
	    // Have not seen the language.  Create it. 
	    languageCounts.put(language, new Integer(1));
	} else {
	    // We have seen the language at least once.  
	    // Now we have seen it again.  Update it. 
	    languageCounts.put(language, new Integer(((Integer)languageCounts.get(language)).intValue() +1));
	}
	// Check to see if this annotator exists
	if (annotatorCounts.get(annotatorID) == null) {
	    // Create a new entry with language language
	    Hashtable newHash = new Hashtable();
	    newHash.put(language, new Integer(1));
	    annotatorCounts.put(annotatorID, newHash);
	} else {
	    // Check to see if this language exists, update
	    Hashtable annHash = (Hashtable) annotatorCounts.get(annotatorID);
	    if (annHash.get(language) == null) {
		// Have not seen the language.  Create it. 
		annHash.put(language, new Integer(1));
	    } else {
		// We have seen the language at least once.  
		// Now we have seen it again.  Update it. 
		annHash.put(language, new Integer(((Integer)annHash.get(language)).intValue() +1));
	    }	    
	}
	// Done.  
    }
    
    /**
     *  getLanguageCount(String lang) 
     *  
     *  Returns the number of annotations that we have seen for the 
     *  language lang, 0 if none
     */
    public int getLanguageCount(String lang) {
	if (languageCounts.get(lang) != null) {
	    return ((Integer)languageCounts.get(lang)).intValue();
	} else {
	    return 0;
	}
    }
    

    /**
     *  getAnnotatorLanguageCount(String annID, String lang)
     *  
     *  Return the number of annotations that we have seen for 
     *  this annotator language pair.  0 if none.
     */
    public int getAnnotatorLanguageCount(String annID, String lang) {
	if (annotatorCounts.get(annID) == null) {
	    return 0;
	} else if (((Hashtable)annotatorCounts.get(annID)).get(lang) == null) {
	    return 0;
	} else {
	    return ((Integer)((Hashtable)annotatorCounts.get(annID)).get(lang)).intValue();
	}
    }

    /**
     *  getAnnotatorCount(String annID) 
     * 
     *  Return the number of annotations that we have seen by
     *  this annotator over all languages done by this annotator.
     *  0 if we have not seen this annotator
     */
    public int getAnnotatorCount(String annID) {
	if (annotatorCounts.get(annID) == null) {
	    return 0;
	} else {
	    int sum=0;
	    Enumeration els = ((Hashtable)annotatorCounts.get(annID)).elements();
	    while (els.hasMoreElements()) {
		sum += ((Integer)els.nextElement()).intValue();
	    }
	    return sum;
	}
    }

    /**
     *  getNumAnnotators()
     *
     *  Return the number of annotators that we have seen
     */
    public int getNumAnnotators() {
	return annotatorCounts.size();
    }

    /**
     *  getNumLanguages() 
     * 
     *  Return the number of languages that we have seen
     */
    public int getNumLanguages() {
	return languageCounts.size();
    }
    
    /**
     *  getNumAnnotatorLanguagePairs() 
     * 
     *  Return the number of annotator language pairs.
     */
    public int getNumAnnotatorLanguagePairs() {
	int sum=0;
	Enumeration ans = annotatorCounts.elements();
	while (ans.hasMoreElements()) {
	    // See how many languages this annotator has done
	    sum += ((Hashtable)ans.nextElement()).size();
	}
	return sum;
    }

    /** 
     *  getLanguages() 
     *
     *  Return an array of strings with that languages that we know.
     */
    String[] getLanguages() {
	int numLangs = getNumLanguages();
	String[] langs = new String[numLangs];
	// Fill in the languages
	Enumeration langsEnum = languageCounts.keys();
	int i=0;
	while (langsEnum.hasMoreElements()) {
	    langs[i++] = (String)langsEnum.nextElement();
	}
	return langs;
    }

    /**
     *  getAnnotators() 
     *
     *  Return an array of strings that are the annotator IDs that we know
     */
    String[] getAnnotators() {
	int numAnns = getNumAnnotators();
	String[] ans = new String[numAnns];
	// Fill in the annotators
	Enumeration ansEnum = annotatorCounts.keys();
	int i=0;
	while (ansEnum.hasMoreElements()) {
	    ans[i++] = (String)ansEnum.nextElement();
	}
	return ans;
    }

    /**
     *  getAnnotatorLanguagePairs() 
     *
     *  Return an array of strings in the form 
     *  AnnotatorID (Language) #of annotations#
     *  for all of the ones that we know about.
     */
    public String[] getAnnotatorLanguagePairs() {
	int total = getNumAnnotatorLanguagePairs();
	String[] ans = new String[total];
	Enumeration anKeys = annotatorCounts.keys();
	int i=0;
	while (anKeys.hasMoreElements()) {
	    // Get all the languages for this annotator
	    String currentAnn = (String)anKeys.nextElement();
 	    Enumeration langs = ((Hashtable)annotatorCounts.get(currentAnn)).keys();
	    while (langs.hasMoreElements()) {
		// Add in a new one for this annotator (language)
		String currLang = (String)langs.nextElement();
		ans[i++] = ""+ currentAnn + " ("+currLang+") " + getAnnotatorLanguageCount(currentAnn, currLang);
	    }
	}
	return ans;
    }
    
    void debugInfo() {
	System.out.println("Annotators: "+annotatorCounts);
	System.out.println("Languages: "+languageCounts);
    }

} // GMAOLanguagePairs



