
/**
 * GMAOServer.java
 *
 *  This class loads up XML files that conform to the Manga.dtd, and
 *  accepts RMI calls to update those DOMs with new information.  
 *
 *  Also keeps track of different annotators, and IPs that are banned,
 *  etc.
 *
 * Created: Mon Jul 24 09:15:28 2000
 *
 * @author devans@cs.columbia.edu
 * @version 0.1b */

import java.util.*;
import java.io.*;
import java.rmi.*;
import java.rmi.server.*;
import org.w3c.dom.*;
import java.lang.reflect.Array;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.DocumentBuilder;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
// Specific sun class we will use to write out the DOM
import com.sun.xml.tree.XmlDocument;


/* 
   GMAO - Image Annotation for Manga Translation
   Copyright (C) 2000  David Kirk Evans
   
   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

   David Kirk Evans (devans@cs.columbia.edu)
   http://www.cs.columbia.edu/~devans/   
*/

public class GMAOServer extends UnicastRemoteObject
    implements GMAORemote {
    
    Properties prefs;      // Properties file for us
    Document[] documents;  // Array to hold all the documents
    String[] files; // List of file names to save the documents to 
    Document annotatorConfig; // Document with Annotators and banned
    Hashtable titleIDMap;  // Maps title ID strings to ints to index documents array
    Hashtable annotators;  // Annotator IDs to passwords
    Hashtable bannedAnnotators;  // Banned Annotator IDs
    DocumentBuilder docBuilder=null; // Keep this guy around for later use

    public GMAOServer() throws java.rmi.RemoteException {
	super();

	/* Open up a properties file for us */
	prefs = new Properties();
	try {
	    prefs.load(new FileInputStream("GMAOServer.props"));
	    System.out.println("Loaded in GMAOServer.props");
	} catch (FileNotFoundException e) {
	    System.out.println("GMAOServer.props does not exist, using defaults.");
	} catch (IOException e) {
	    System.out.println("Error loading GMAOServer.props: "+e);	    
	}
	titleIDMap = new Hashtable();
	annotators = new Hashtable();
	bannedAnnotators = new Hashtable();
	/** 
	 *  Look in the properties file for the XML documents that
	 *  we are supposed to keep open and update.  These will be
	 *  listed in the properties file as something like 
	 *  xmlFile.0  path to xml file
	 *  xmlFile.1  path to xml file 
	 *    etc...
	 */	 
	// First, count how many there are
	int i = 0;
	while (prefs.getProperty("xmlFile."+i) != null) {
	    i++;
	}
	System.out.println("Found "+i+" XML files to read");
	// Create a new Array of Documents 
	documents = (org.w3c.dom.Document[]) Array.newInstance(org.w3c.dom.Document.class, i);
	// Create an array of file for these docs too
	files = (String[]) Array.newInstance(String.class, i);
	// Read in all of the XML files.
	try {
	    DocumentBuilderFactory 
		docBuilderFactory = DocumentBuilderFactory.newInstance();
	    docBuilder = docBuilderFactory.newDocumentBuilder();
	    
	    for (int j=0; j<i; j++) {
		System.out.println ("Reading in XML file "+prefs.getProperty("xmlFile."+j));
		files[j] = (String)prefs.getProperty("xmlFile."+j);
		documents[j] = docBuilder.parse(new File(files[j]));
		// normalize text representation
		documents[j].getDocumentElement().normalize();
		System.out.println ("Read in XML file "+prefs.getProperty("xmlFile."+j));
		// Set up a hastable that maps the string of
		// SeriesTitle and SeriesNumber to an int
		// Get the Node BookTitle
		String docID = "";
		NodeList tempL = documents[j].getDocumentElement().getElementsByTagName("SeriesName");
		docID += tempL.item(0).getFirstChild().getNodeValue();
		tempL = documents[j].getDocumentElement().getElementsByTagName("SeriesNumber");
		docID += tempL.item(0).getFirstChild().getNodeValue();
		//  Now should have "SeriesName SeriesNumber" 		
		titleIDMap.put(docID, new Integer(j));
		System.out.println("Associated "+docID+" to index "+j);
	    }
	    // Read in the XML file with annotator IDs, names, and passwords
	    annotatorConfig = docBuilder.parse(new File(prefs.getProperty("annotatorConfigFile")));
	    System.out.println("Read in annotator config file "+prefs.getProperty("annotatorConfigFile"));	    
	    // Set up Annotator table
	    NodeList annList = annotatorConfig.getElementsByTagName("Annotator");
	    for (int j=0; j<annList.getLength(); j++) {
		NamedNodeMap attr = annList.item(j).getAttributes();
		Annotator temp = new Annotator(attr.getNamedItem("AnnID").getNodeValue(), attr.getNamedItem("Name").getNodeValue(), attr.getNamedItem("Password").getNodeValue());
		// Associate the Annotator temp with the id
		annotators.put(attr.getNamedItem("AnnID").getNodeValue(), temp);
		System.out.println("Added Annotator "+temp.getID());
	    }
	    // Set up banned Annotator table
	    annList = annotatorConfig.getElementsByTagName("Banned");
	    for (int j=0; j<annList.getLength(); j++) {
		NamedNodeMap attr = annList.item(j).getAttributes();
		Node id = attr.getNamedItem("AnnID");
		if (id != null) {
		    bannedAnnotators.put(id.getNodeValue(), new Boolean(true));
		    System.out.println("Banned Annotator "+id.getNodeValue());
		}
		id = attr.getNamedItem("IP");
		if (id != null) {
		    bannedAnnotators.put(id.getNodeValue(), new Boolean(true));
		    System.out.println("Banned IP "+id.getNodeValue());
		}
	    }
	    
	} catch (SAXParseException err) {
	    System.out.println ("** Parsing error" 
				+ ", line " + err.getLineNumber ()
				+ ", uri " + err.getSystemId ());
	    System.out.println("   " + err.getMessage ());
	    // print stack trace as below
	} catch (SAXException e) {
	    Exception	x = e.getException ();
	    
	    ((x == null) ? e : x).printStackTrace ();
	    
	} catch (Throwable t) {
	    t.printStackTrace ();
	}
    }
  
    /**
     *  This method logs an Annotator in.  If the Annotator already
     *  exists, and the password does not match, then false is
     *  returned.  If the Annotator exists, and the password matches,
     *  true is returned.  Or if the Annotator does not exist, a new
     *  one is created and their password is recorded.  
     *
     *  Throws GMAOAnnotatorBannedException if the Annotator has been
     *  banned on this server.
     */
    public boolean logInAnnotator(Annotator a) throws RemoteException, GMAOAnnotatorBannedException {
	// Check to see if this annotator is banned
	if (bannedAnnotators.get(a.getID()) != null) {
	    throw new GMAOAnnotatorBannedException("Annotator "+a.getID()+" has been banned on this server.");
	}
	//  Check for IP banning here.
	String remote = "unknown";
	try {
	    remote = java.rmi.server.RemoteServer.getClientHost();
	    if (bannedAnnotators.get(remote) != null) {
		throw new GMAOAnnotatorBannedException("IP "+remote+" has been banned on this server.");
	    }
	} catch (java.rmi.server.ServerNotActiveException snae) {
	    System.out.println("Error trying to ID remote client: "+snae);
	}
	
	Annotator local = (Annotator) annotators.get(a.getID());
	if (local != null) {
	    // We have an annotator by this ID.  Check to see if 
	    // the name and passwords match
	    if (local.getPassword().equals(a.getPassword()) &&
		local.getName().equals(a.getName())) {
		// Sucessuful login
		System.out.println(remote + " " + local.getID() + " logged in");
		return true;
	    } else {
		System.out.println(remote + " " + local.getID() + " tried to log in, but failed");
		return false;
	    }
	} else {
	    // We have a new annotator.  Put him in.
	    annotators.put(a.getID(), a);
	    // Also put into the XML file
	    try {
		Element na = annotatorConfig.createElement("Annotator");
		na.setAttribute("AnnID", a.getID());
		na.setAttribute("Name", a.getName());
		na.setAttribute("Password", a.getPassword());
		annotatorConfig.getDocumentElement().appendChild(na);
		// Save the annotatorConfig file
		saveDom(annotatorConfig, prefs.getProperty("annotatorConfigFile"));
		System.out.println(remote + " created new Annotator: "+a.getID());
	    } catch (DOMException de) {
		System.out.println("DOMException while trying to create new Annotator: "+de);
	    }
	    return true;
	}
    }
  
    /**
     *  This method will update the DOM in the server using the the
     *  string representation of a DocumentFragment that has just the
     *  page information and the bubbles that have changed since the
     *  page was loaded on the client.  The server will compare the
     *  two DOMs and insert only the new nodes that exist in d and not
     *  in the Server DOM.  The comparison is based on the text of the
     *  Translation bubbles, not the bubble number.  To make sure the
     *  bubbles are the same, it just looks at the location of the
     *  bubble.
     *
     *  The docID string identifies the XML document that we are
     *  working on.  It is a combination of the XML fields series
     *  title and series number.
     *
     *  Synchronized because we only want one call to this method at a
     *  time.  Right?  -- testing right now with not being
     *  synchronized because I think that might be causing my machine
     *  to freeze...
     *
     *  Huge problem right now is that we are unable to create new chapters 
     *  using the current implementation. 
     *
     *  Here is what is important (in the server side really) :
     *  m_Doc.load(new BufferedInputStream(new ByteArrayInputStream(xmlStr.getBytes()))); */
    public void updateDOM(String docID, String changedXMLString) throws RemoteException {
	// System.out.println("Trying to change page with: "+changedXMLString);
	// Parse the string into a DOM
	if (docBuilder == null) {
	    System.out.println("docBuilder is null: I can't parse without it!");
	    return;
	}
	org.w3c.dom.Document updates=null;
	String remote = "unknown";
	try {
	    remote = java.rmi.server.RemoteServer.getClientHost();
	} catch (java.rmi.server.ServerNotActiveException snae) {
	    System.out.println("Error trying to ID remote client: "+snae);
	}

	try {
	    // Questions here about what encoding format string should be in
	    /**
	     *  Have hard-coded euc-jp character encoding into string here!
	     */
	    updates = docBuilder.parse(new BufferedInputStream(new ByteArrayInputStream(changedXMLString.getBytes("euc-jp"))));
	    // normalize text representation
	    updates.getDocumentElement().normalize();
	    // Locate which dom we are dealing with
	    Document doc = documents[((Integer)titleIDMap.get(docID)).intValue()];
	    // Locate corresponding page in main DOM
	    NodeList pages = doc.getDocumentElement().getElementsByTagName("Page");
	    Element page=null;
	    int i=0;
	    int targetPage=-1;
	    try {
		targetPage = Integer.parseInt(updates.getDocumentElement().getElementsByTagName("Page").item(0).getAttributes().getNamedItem("Number").getNodeValue());
	    } catch (NumberFormatException nfe) {
		System.out.println("Wasn't able to get page numbers... "+nfe);
	    }
	    while (i < pages.getLength()) {
		// Search for the same page number
		page = (Element)pages.item(i);
		if (targetPage == Integer.parseInt(page.getAttributes().getNamedItem("Number").getNodeValue())) {
		    // Found the right page.  Set i = pages.getLength()
		    i = pages.getLength();
		    // System.out.println("Found page "+targetPage);
		}
		i++;
	    }
	    if (i == pages.getLength()) {
		// This page does not exist!  We can just add it.  yay.
		/** 
		 * For right now, new pages always just go at the end
		 * of the current DOM (ie, in the last chapter.)
		 */
		try {
		    /**
		     * Try to change the owner of the node using SUN classes...  This is also
		     * sun implementation dependent!
		     */
		    // System.out.println("About to append "+updates.getDocumentElement().getFirstChild() + "\n to "+doc.getDocumentElement().getLastChild());
		    Node toAdd = updates.getDocumentElement().getFirstChild();
		    try {
			((com.sun.xml.tree.XmlDocument)doc).changeNodeOwner(toAdd);
		    } catch (DOMException f) {
			System.out.println("Exception when trying to change node owner: "+f);
		    }
		    //System.out.println("Changed node owner.");
		    //System.out.println("About to append "+toAdd + "\n to "+doc.getDocumentElement());
		    doc.getDocumentElement().getLastChild().appendChild(toAdd);
		    System.out.println(remote +" added new page: "+toAdd);
		} catch (Exception e) {
		    System.out.println("Error trying to append new page to DOM: "+e);
		}
	    } else {
		/**
		 *  We are checking here to see if two bubbles are
		 *  the same by looking at the point lists for the 
		 *  bubbles.  If those are the same, we are calling
		 *  the bubbles the same.  This has, obviously, some
		 *  problems, but it should be "good enough."  We 
		 *  certainly can't count on the bubble numbering being
		 *  the same.
		 */
		// Check each Bubble to see if it exists in main dom
		NodeList updateBubbles = updates.getDocumentElement().getElementsByTagName("Bubble");
		NodeList currentBubbles = page.getElementsByTagName("Bubble");
		for (i=0; i<updateBubbles.getLength(); i++) {
		    int j=0;
		    while (j<currentBubbles.getLength() && 
			   ! sameBubble((Element)currentBubbles.item(j), (Element)updateBubbles.item(i))) {
			j++;
		    }
		    if (j == currentBubbles.getLength()) {
			// It is a new bubble entirely.  Add it
			// Change owner of toAdd
			Node toAddNode = updateBubbles.item(i);
			try {
			    ((com.sun.xml.tree.XmlDocument)doc).changeNodeOwner(toAddNode);
			} catch (DOMException f) {
			    System.out.println("Exception when trying to change node owner: "+f);
			}
			// Change the "Number" attribute on the bubble to be the last bubble in current Page
			((Element)toAddNode).setAttribute("Number", ""+j);
			page.appendChild(toAddNode);
			// Because we are moving the node out of currentBubbles
			// we have to keep the counter the same.  Strange, I know.
			i--;
			System.out.println(remote + " added new bubble: "+toAddNode);
			// System.out.println("Now page is: "+page);
		    } else {
			// Bubble exists, check to see which notes changed
			/** 
			 * Since there is no way to delete a note, 
			 * the order on them should be stable. 
			 * Use the order to compare for new notes/edits.
			 */
			
			// System.out.println("Updating bubble "+j);
			
			NodeList updateTrans = ((Element)updateBubbles.item(i)).getElementsByTagName("Translation");
			NodeList currentTrans = ((Element)currentBubbles.item(j)).getElementsByTagName("Translation");			
			int k=0; // Use k to step through the currentBubbles
			while (updateTrans.getLength() > 0) {
			    // System.out.println("k is now "+k+" currentTrans.getLength() "+currentTrans.getLength()+" updateTrans author: "+updateTrans.item(0).getAttributes().getNamedItem("Author").getNodeValue());
			    // Change note only if annotator ID is the same
			    if (k < currentTrans.getLength() && updateTrans.item(0).getAttributes().getNamedItem("Author").getNodeValue().equals(currentTrans.item(k).getAttributes().getNamedItem("Author").getNodeValue())) {
				// The annotators are the same, replace them
				Node toAddNode = updateTrans.item(0);
				// Change owner of toAdd
				System.out.println(remote+" apdated "+currentTrans.item(k)+" to "+toAddNode);
				try {
				    ((com.sun.xml.tree.XmlDocument)doc).changeNodeOwner(toAddNode);
				} catch (DOMException f) {
				    System.out.println("Exception when trying to change node owner: "+f);
				}				
				currentBubbles.item(j).replaceChild(toAddNode, currentTrans.item(k));
			    } else {
				// Or if it is a new note, add it.
				Node toAddNode = updateTrans.item(0);
				System.out.println(remote+" added new translation "+toAddNode);
				try {
				    ((com.sun.xml.tree.XmlDocument)doc).changeNodeOwner(toAddNode);
				} catch (DOMException f) {
				    System.out.println("Exception when trying to change node owner: "+f);
				}				
				currentBubbles.item(j).appendChild(toAddNode);
				// System.out.println("Now bubble is: "+currentBubbles.item(j));
			    }
			    k++;
			}
		    }    
		}
	    }	    
	    // Save out the DOM
	    saveDom(doc, files[((Integer)titleIDMap.get(docID)).intValue()]);
	} catch(DOMException de) {
	    System.out.println("Error trying to parse update XML: "+de);
	} catch (NumberFormatException nf) {
	    System.out.println("Error trying to get page number from update XML: "+nf);
	} catch (IOException ioe) {
	    System.out.println("IO Exception while trying to parse updates: "+ioe);
	} catch (SAXException se) {
	    System.out.println("SAXException while trying to parse updates: "+se);
	}
    }

    /**
     *  Saves the document doc to the file file.
     *  Depends on using the xml Sun classes.  
     *  Also, hard-coded right here to "euc-jp" encoding system.
     */
    public void saveDom(org.w3c.dom.Document doc, String filename) {
	// System.out.println("Saving "+doc.getDocumentElement());
	// Cast our mangaDoc to a specific sun implementation.
	XmlDocument xdoc = (XmlDocument) doc;
	// Write it out to a File
	try {
	    File tempFile = File.createTempFile("GMS_", ".xml");
	    System.out.println("About to write "+tempFile.toString());
	    FileOutputStream tmpStream = new FileOutputStream(tempFile);
	    xdoc.write(new OutputStreamWriter(tmpStream, "euc-jp"));
	    // Now, try to make sure that the file is really written - finalize it
	    tmpStream.close();
	    // Write must have been successful
	    // Move old file to a backup file name
	    File file = new File(filename);
	    // Delete old backup
	    File oldBackup = new File(filename+".bak");
	    if (!oldBackup.delete()) {
		System.out.println("Error deleting old backup file "+filename+".bak");
	    }
	    oldBackup = null;
	    if (!file.renameTo(new File(filename+".bak"))) {
		System.out.println("Error renaming current file "+filename+" to "+filename+".bak");
	    }	    
	    // Re-name to the original file name
	    if (!tempFile.renameTo(new File(filename))) {
		System.out.println("Error renaming "+tempFile.toString()+" to "+filename);
	    } else {
		System.out.println("Saved file "+filename);
	    }
	} catch (java.io.IOException e) {
	    System.out.println("Couldn't write out the DOM, file io error: "+e);
	}
	
    }

    /**
     *  public boolean sameBubble(b1, b2)
     *  Given two org.w3c.dom.Element objects, we will return
     *  a boolean set to true if we think that they are the 
     *  same bubble.  Or at least close enough to be.  
     *  We do the comparison based on the points in the shapes
     *  for the bubbles.  If the points are the same, they are the
     *  same bubble.  That is close enough.
     */
    public boolean sameBubble(Element b1, Element b2) {
	NodeList p1 = b1.getElementsByTagName("Point");
	NodeList p2 = b2.getElementsByTagName("Point");
	// System.out.println("Length pt1: "+p1.getLength()+ " length pt2: "+p2.getLength());
	if (p1.getLength() != p2.getLength()) {
	    // Got to have the same number of points.
	    return false;
	}
	int i=0;
	int p1x, p1y, p2x, p2y;
	while (i<p1.getLength()) {
	    // Get X1, Y1 for p1
	    try {
		p1x = Integer.parseInt(p1.item(i).getAttributes().getNamedItem("X").getNodeValue());
		p1y = Integer.parseInt(p1.item(i).getAttributes().getNamedItem("Y").getNodeValue());
		p2x = Integer.parseInt(p2.item(i).getAttributes().getNamedItem("X").getNodeValue());
		p2y = Integer.parseInt(p2.item(i).getAttributes().getNamedItem("Y").getNodeValue());
		if (p1x != p2x || p1y != p2y) {
		    // System.out.println("Point ("+p1x+", "+p1y+") != ("+p2x+", "+p2y+")");
		    // Not the same
		    return false;
		}		
	    } catch (NumberFormatException nfe) {
		System.out.println("Couldn't parse number when comparing bubbles! "+nfe);
	    }
	    i++;
	}
	// The same! WOW!
	return true;
    }

    /**
     *  This method is used to indicate that a new chapter has been 
     *  added to the manga.  All subsequent added pages will go into 
     *  this chapter.
     */
    public void addChapter(String docID, int number, String title) throws RemoteException {
	Document mangaDoc = documents[((Integer)titleIDMap.get(docID)).intValue()];
	String remote = "unknown";
	try {
	    remote = java.rmi.server.RemoteServer.getClientHost();
	} catch (java.rmi.server.ServerNotActiveException snae) {
	    System.out.println("Error trying to ID remote client: "+snae);
	}

	if (mangaDoc == null) {
	    System.out.println(remote+" tried to add a new Chapter "+number+" : "+title+" to non-existant documentID "+docID);
	    return;
	}
	// Add a new Chapter.  Just make the new node in the DOM.
	// The first node is the MangaTranslation, add under that.
	org.w3c.dom.Element newChap = mangaDoc.createElement("Chapter");
	// Add attribute "Number" to the new chapter
	newChap.setAttribute("Number", ""+number);
	// Add a new node, ChapterTitle, with one element, the text title
	org.w3c.dom.Element chapTitle = mangaDoc.createElement("ChapterTitle");
	chapTitle.appendChild(mangaDoc.createTextNode(title));
	newChap.appendChild(chapTitle);
	/* Why is this lastchild?? You ask?  I think it is because 
	   the first node is actually a node that tells what type of
	   object the document is.  The second (and last) node is 
	   actually the first child, the MangaTranslation.
	*/
	mangaDoc.getLastChild().appendChild(newChap);
	System.out.println(remote+" added new Chapter "+number+" : "+title);
	// Save out the DOM
	saveDom(mangaDoc, files[((Integer)titleIDMap.get(docID)).intValue()]);
    }

    /**
     *  Set up an instance of the server, bind it to the rmiregistry
     */
    public static void main(String[] args) {
        /* 
	   if (System.getSecurityManager() == null) {
	   System.setSecurityManager(new RMISecurityManager());
	   }
	*/
        String name = "/GMAOServer";
        try {
            GMAOServer mangaServer = new GMAOServer();
            Naming.rebind(name, mangaServer);
            System.out.println("GMAOServer bound in registry on local host");
        } catch (Exception e) {
            System.err.println("GMAOServer exception: " + 
			       e.getMessage());
            e.printStackTrace();
        }
    }


} // GMAOServer
