
/**
 * Note.java
 *
 *  This class implements a note that is part of an annotation.
 *  Annotations cover and area on an image, the notes are specific
 *  pieces of text information associated with that area of the image.
 *  The notes in particular have these fields: the String note itself,
 *  an Annotator for the person that annotated the location with this
 *  note, and a date for when the annotation was created.
 *
 * For the future: 
 *
 *  - Let Notes have user-defined fields as well as the three fields
 *  described above.  The fields would be defined when the Note is
 *  created by passing information regarding what the possible field
 *  names are, default values, and whether or not the field is
 *  required or not.  When creating a new Note, the required fields
 *  will be checked for existence, and it will be fine to leave the
 *  optional fields blank.  
 *
 *  - Add a method to a Note that returns a JPanel that displays the
 *  content of the Note in a read-only display (that formats the
 *  required fields and the user fields nicely.)  This will allow us
 *  to just place this returned JPanel somewhere to display the Note.
 *
 *  - For new Note creation - this will likely be done in
 *  Annotation.java - when the user tries to create a new Note, pop up
 *  a modal dialog box that prompts them for the necessary information
 *  (the actual text of the Note, and whatever user field values are
 *  required.)  This will take care of validation, and clean up the
 *  confusion in the interface.  Can use the same box to "Edit" an
 *  existing Note.
 *
 * Created: Fri Mar 17 16:11:12 2000
 *
 * @author devans
 * @version 1.0 */

/* 
   GMAO - Image Annotation for Manga Translation
   Copyright (C) 2000  David Kirk Evans
   
   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

   David Kirk Evans (devans@cs.columbia.edu)
   http://www.cs.columbia.edu/~devans/   
*/

import java.util.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import java.text.DateFormat;

//  GMAO2K imports 
import NoteCustomField;

public class Note implements java.io.Serializable {
    String theNote;  // The info for the note itself.
    Annotator theAnnotator;  //  Who annotated this note?
    Date theDate;  // When was this note made?
    Vector customFields;  // The custom fields for this Note
    int number;  //  The number of this note

    /**
     *  Create a new Note that has the String n as the note information,
     *  set the time this was created to the current time, and have no
     *  custom fields defined for this Note.
     */
    public Note(String n, int num) {
	theNote = n;
	// Set the date
	theDate = new Date();
	customFields = new Vector();  //  empty vector though. 
	number = num;
    }

    /**
     *  Create a new Note that has the String n as the note information, 
     *  set the time this was created to the current time, and has the 
     *  custom fields stored in the Vector customFieldsVector.  Note 
     *  that customFieldsVector should be full to the brim with objects
     *  of the class NoteCustomField
     */
    public Note(String n, int num, Vector customFieldsVector) {
	this(n, num);  // call the constructor with only a String
	// Point to the right customFields vector
	customFields = customFieldsVector;
    }
    
    public void setNote(String s) {
	theNote = s;
    }
    public void setAnnotator(Annotator a) {
	theAnnotator = a;
    }
    public String getNote() {
	return theNote;
    }
    public int getNumber() {
	return number;
    }
    public Date getDate() {
	return theDate;
    }
    public void setDate(Date d) {
	theDate = d;
    }
    public Annotator getAnnotator() {
	return theAnnotator;
    }
    public Vector getCustomFields() {
	return customFields;
    }
    public void addCustomField(NoteCustomField cf) {
	customFields.add(cf);
    }
    /**
     * Given a string of the name of a custom field, return its value.
     * Null if no such field found, the string value otherwise.
     */
    public String getCustomFieldValue(String field) {
	for (int i=0; i<customFields.size(); i++) {
	    NoteCustomField ncf = (NoteCustomField) customFields.elementAt(i);
	    if (ncf.getName().equals(field)) {
		return ncf.getValue();
	    }
	}
	return null;
    }

    public void setCustomFieldValue(String field, String value) {
	for (int i=0; i<customFields.size(); i++) {
	    NoteCustomField ncf = (NoteCustomField) customFields.elementAt(i);
	    if (ncf.getName().equals(field)) {
		ncf.setValue(value);
		return;
	    }
	}
	// Didn't find it, create a new NoteCustomField for it
	addCustomField(new NoteCustomField(field, value, true));
    }
    
    /**
     *  Return a text string that is HTML that represents the 
     *  content of this note.
     */
    public String getHTMLRepresentation() {
	String ans = "Note # "+number+" Created: " + DateFormat.getDateInstance().format(theDate)+"<BR>\n";
	//  Check to see if we have any customfields
	for (int i=0; i<customFields.size(); i++) {
	    //  Add in custom field at i
	    NoteCustomField field = (NoteCustomField) customFields.elementAt(i);
	    //  Only add this one if it has a value
	    if (field.getValue() != null && ! field.getValue().equals("")) {
		ans += field.getName()+": "+field.getValue()+"<BR>";
	    }
	}
	ans += "<HR>"+ theNote;
	return ans;
    }
    
    /**
     *  Returns a JPanel that shows a representation of the
     *  information in this note.  This includes the custom fields
     *  that this Note might have.  The JPanel returned is suitable
     *  for stashing in some displayed area.  It is only for display
     *  though.  Custom fields that do not have a value are not
     *  displayed.  
     */
    public JPanel getDisplayPanel() {
	JPanel me = new JPanel();
	
	//  At any rate, we want to display the note number and the
	//  text of the note
	
	//  Set the layout Manager for me to be a borderlayout
	me.setLayout(new BorderLayout());
	//  Add to the top a BoxLayout with Note and Date
	Box temp = new Box(BoxLayout.X_AXIS);
	temp.add(new JLabel("Note #"+number));
	//  Put some space in between these two.
	temp.add(Box.createHorizontalGlue());
	temp.add(new JLabel("Created: " + DateFormat.getDateInstance().format(theDate)));
	//  Add the number/date panel to me in the NORTH
	me.add(temp, BorderLayout.NORTH);
	//  Now the harder part: The center should be any custome fields
	//  that are defined, taking up minimum vertical space.  Split
	//  that with a big text area for the actual annotation.  
	Box bigTemp = new Box(BoxLayout.Y_AXIS);
	//  Check to see if we have any customfields
	for (int i=0; i<customFields.size(); i++) {
	    //  Add in custom field at i
	    NoteCustomField field = (NoteCustomField) customFields.elementAt(i);
	    //  Only add this one if it has a value
	    if (field.getValue() != null && ! field.getValue().equals("")) {
		temp = new Box(BoxLayout.X_AXIS);
		temp.add(new JLabel(field.getName()+":"));
		temp.add(new JLabel(field.getValue()));
		//  Add it in
		bigTemp.add(temp);
	    }
	}
	//  Add in the text area for the note
	// System.out.println("Showing note "+theNote);
	JTextArea temptext = new JTextArea(theNote);
	temptext.setLineWrap(true);   // wrap words
	temptext.setWrapStyleWord(true);
	temptext.setEditable(false);  // don't let people edit it
	bigTemp.add(new JScrollPane(temptext));  // put is in a scrollpane
	//  Now add in the bigTemp to me
	me.add(bigTemp);
	//  Reserve the SOUTH for something like Annotator information 
	me.revalidate();
	return me;
    }
    
} // Note





