
/**
 * XMLTreeDisplay.java
 *
 *  This class will display a split pane with a JTree of the Manga 
 *  XML DOM on the left, and the AnnotationPanel on the right.  
 *
 * Created: Sun Jun 25 15:48:52 2000
 *
 * @author David Evans
 * @version
 */

/* 
   GMAO - Image Annotation for Manga Translation
   Copyright (C) 2000  David Kirk Evans
   
   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

   David Kirk Evans (devans@cs.columbia.edu)
   http://www.cs.columbia.edu/~devans/   
*/

// Document for the XML Doc
import org.w3c.dom.Document;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.DocumentBuilder;

// GUI components and layouts
import javax.swing.JFrame;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTree; 
import javax.swing.JSplitPane;
import javax.swing.JEditorPane;

// GMAO stuff
import AnnotatePanel;

// GUI support classes
import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.Toolkit;
import java.awt.event.WindowEvent;
import java.awt.event.WindowAdapter; 

// For creating borders
import javax.swing.border.EmptyBorder;
import javax.swing.border.BevelBorder;
import javax.swing.border.CompoundBorder; 

// For creating a TreeModel
import javax.swing.tree.*;
import javax.swing.event.*;
import java.util.*;


public class XMLTreeDisplay extends JPanel {
    Document document = null; 
    JTree tree;  
    Properties props; 
    AnnotatePanel annPane;
    org.w3c.dom.Node lastSelection = null; // Last selected node in JTree
    GMAORemote server=null;  // The remote server we will use
    String docID="";  // Identifier for doc we are currently working with
    boolean annotatorLoggedIn=false; // Boolean to indicate if annotator is logged in - if not, don't send updates of any kind to the server!!
    static final int windowHeight = 460;
    static final int leftWidth = 300;
    static final int rightWidth = 340;
    static final int windowWidth = leftWidth + rightWidth;     

    // An array of names for DOM node-types
    static final String[] typeName = {
        "none",
        "Element",
        "Attr",
        "Text",
        "CDATA",
        "EntityRef",
        "Entity",
        "ProcInstr",
        "Comment",
        "Document",
        "DocType",
        "DocFragment",
        "Notation",
    };
    /* Contants to identify DOM node types */    
    static final int ELEMENT_TYPE =   1;
    static final int ATTR_TYPE =      2;
    static final int TEXT_TYPE =      3;
    static final int CDATA_TYPE =     4;
    static final int ENTITYREF_TYPE = 5;
    static final int ENTITY_TYPE =    6;
    static final int PROCINSTR_TYPE = 7;
    static final int COMMENT_TYPE =   8;
    static final int DOCUMENT_TYPE =  9;
    static final int DOCTYPE_TYPE =  10;
    static final int DOCFRAG_TYPE =  11;
    static final int NOTATION_TYPE = 12; 

    // The list of elements to display in the tree if we are 
    // compressing the view.  Check properties to see if we are. 
    static String[] treeElementNames = {
	"MangaTranslation",
        "BookTitle",
        "Chapter",
        "Page",
	"Bubble"
    };

    /**
     *  This sets the remote server that we will use for updates.
     *  If it is null, we will not do any updates.  If it is not null,
     *  then we will update.  
     */
    public void setServer(GMAORemote serv) {
	server = serv;
    }

    /**
     *  Sets the document identifier string up for remote updates
     */
    public void setDocID(String d) {
	docID = d;
    }
    
    boolean treeElement(String elementName) {
	for (int i=0; i<treeElementNames.length; i++) {
	    if (elementName.equals(treeElementNames[i])) 
		return true;
	}
	return false;
    } 
    
    public XMLTreeDisplay(Properties settings) {
	props = settings;
	// Make a nice border
	EmptyBorder eb = new EmptyBorder(5,5,5,5);
	BevelBorder bb = new BevelBorder(BevelBorder.LOWERED);
	CompoundBorder cb = new CompoundBorder(eb,bb);
	this.setBorder(new CompoundBorder(cb,eb));
	
	// Set up the tree
	tree = new JTree(new DomToTreeModelAdapter());
	
	// Build left-side view
	JScrollPane treeView = new JScrollPane(tree);
	treeView.setPreferredSize(new Dimension( leftWidth, windowHeight ));
	
	// Build right-side view
	annPane = new AnnotatePanel(props);
	// JScrollPane annView = new JScrollPane(annPane);
	// annView.setPreferredSize(new Dimension( rightWidth, windowHeight ));
	
	// Build split-pane view
	JSplitPane splitPane = new JSplitPane( JSplitPane.HORIZONTAL_SPLIT,
					       treeView,
					       annPane );
	splitPane.setContinuousLayout( true );
	splitPane.setDividerLocation( leftWidth );
	splitPane.setPreferredSize(new Dimension( windowWidth + 10, windowHeight+10 ));
	//  Add a tree listener to the tree to track selections 
	tree.addTreeSelectionListener(
         new TreeSelectionListener() {
	    public void valueChanged(TreeSelectionEvent e) {
		TreePath p = e.getNewLeadSelectionPath();
		if (p != null) {
		    AdapterNode adpNode = 
			(AdapterNode) p.getLastPathComponent();
		    // Change the contents of the AnnotatePanel here
		    // Get the name of the image to place in there.
		    if (adpNode.nodeName().equals("Page")) {
			/* We have a page selected.  */
			/* Check to see if we had a different page selected 
			   before.  If so, generate the documentFragment for 
			   the currently displayed page, and add it into the
			   DOM.  Then we can change the page displayed. */
			updateDom();
			// Make sure that there is an annotator logged in before we send *any* changes
			// This prevents unauthorized people from sending in new pages and chapters 
			if (server != null && annotatorLoggedIn == true) {
			    try {
				// Create a document fragment that lists the bubbles that have changed.  
				org.w3c.dom.DocumentFragment dfc = annPane.createDomChanges(document);
				// If this has any differences...
				int pageNum=0;
				if (lastSelection != null && dfc != null) {
				    // Get the page number information from lastSelection				    
				    /**
				     *  I wanted to just do something
				     *  simple like send over the DOM.  I
				     *  couldn't do that because the DOM
				     *  isn't serializable.  Then I
				     *  decided to just send over a vector
				     *  of Annotations.  I couldn't do
				     *  that because the Shape member of
				     *  an Annotation is not serializable.
				     *  So now I have decided to send over
				     *  a string, and re-parse it at the
				     *  server.
				     **/
				    // Clone the page node from lastSelection - not DEEP though, only want attributes..
				    org.w3c.dom.Node pageClone = lastSelection.cloneNode(false);
				    // Add in the children from the changed dom...
				    pageClone.appendChild(dfc);
				    // Turn that pageClone - now it has the page info and updates - into a string
				    /**
				     *  Note: this is the second place that I have a dependency on the sun parser.
				     *  I use it to create an XML string here.
				     */
				    String changedXMLString = "<?xml version=\"1.0\" encoding=\"EUC_JP\"?>\n<!DOCTYPE PageFragment [\n";
				    changedXMLString += "<!ELEMENT PageFragment (Page)>\n";
				    changedXMLString += "<!ELEMENT Page (Bubble*)>\n";
				    changedXMLString += "<!ATTLIST Page Number CDATA #REQUIRED>\n<!ATTLIST Page Image CDATA #REQUIRED>\n";
				    changedXMLString += "<!ELEMENT Bubble (Point, Point+, Translation*)>\n";
				    changedXMLString += "<!ATTLIST Bubble Number CDATA #REQUIRED>\n";
				    changedXMLString += "<!ATTLIST Bubble Type (Rectangle|Oval|Polygon) #REQUIRED>\n";
				    changedXMLString += "<!ATTLIST Bubble Character CDATA #REQUIRED>\n";
				    changedXMLString += "<!ELEMENT Point EMPTY>\n";
				    changedXMLString += "<!ATTLIST Point X CDATA #REQUIRED>\n";
				    changedXMLString += "<!ATTLIST Point Y CDATA #REQUIRED>\n";
				    changedXMLString += "<!ELEMENT Translation (#PCDATA)>\n";
				    changedXMLString += "<!ATTLIST Translation Language CDATA #REQUIRED>\n";
				    changedXMLString += "<!ATTLIST Translation Author CDATA #REQUIRED>\n";
				    changedXMLString += "<!ATTLIST Translation Created CDATA #IMPLIED>\n";
				    changedXMLString += "]>\n";
				    changedXMLString += "<PageFragment>\n";
				    changedXMLString += ((com.sun.xml.tree.ElementNode)pageClone).toString();
				    // Update the remote server with the string.  Yeah baby.  Round.  About.  Now.  
				    changedXMLString += "</PageFragment>\n";
				    server.updateDOM(docID, changedXMLString);
				    System.out.println("Sent a change over to the server.");
				} else {
				    // System.out.println("Nothing to update: either lastSelection was null, or there were no changes to update");
				}
				AnnotatePanel.infoDisplay.setText("Updated remote DOM!");
			    } catch (java.rmi.RemoteException re) {
				System.out.println("Remote exception while trying to update remote DOM: "+re);
				AnnotatePanel.infoDisplay.setText("Remote exception while trying to update remote DOM!");
			    }
			}			
			// Set the last selected node to us
			lastSelection = adpNode.getDOMNode();
			/* Send the "Page" node over as a document 
			   fragment to the AnnotatePanel */	
			org.w3c.dom.DocumentFragment df = document.createDocumentFragment();
			//  Must append a CLONE of the node, or the node is MOVED OUT OF OUR DOM!!
			//  We specify a deep clone since we want the children and such. 
			df.appendChild(adpNode.getDOMNode().cloneNode(true));
			// Send the doc fragment over to the AnnotatePanel
			annPane.displayNewPage(props, df);
		    }
		}
	    }
	}
	 );
	
	// Add GUI components
	this.setLayout(new BorderLayout());
	this.add("Center", splitPane );
    }

    public void setProperties(Properties p) {
	props = p;
	// Send it on down
	annPane.setProperties(props);
    }
    
    /**
     *  Pass on the annInfoPanel down to annPane
     */
    public void setAnnInfoPanel(AnnotationInformationPanel aip) {
	annPane.setAnnInfoPanel(aip);
    }
    
    /**
     * updateDom()
     *  Called when we should update the DOM by looking at the 
     *  page, and turning it into a docfragment, then adding that
     *  to the dom in the proper place.  
     *
     *  Generally called when the userselects a new page to view,
     *  also called before saving the file out to make sure that the 
     *  dom is up-to-date.
     */ 
    public void updateDom() {
	if (lastSelection != null) {
	    org.w3c.dom.DocumentFragment bubbles =
		annPane.createDom(document);
	    if (bubbles != null) {
		// Remove old bubbles from lastSelection, put in new ones.
		org.w3c.dom.NodeList oldChildren = lastSelection.getChildNodes();
		org.w3c.dom.NodeList newChildren = bubbles.getChildNodes();
		int i;
		/* Remove all of the children from oldChildren
		   and then add all of the children from newChildren
		*/
		while (oldChildren.getLength() > 0) {
		    lastSelection.removeChild(oldChildren.item(0));
		}
		// Add in all the newChildren -- add in the 
		// doc fragment will actually just add all the children
		lastSelection.appendChild(bubbles);
	    }
	} // We have now swapped out the children on lastSelection
    }
    
    /* If we want to change the model (opened a new XML file, etc.)
     * call this function */
    public void setDOM(Document DOM) {
	document = DOM;
	tree.setModel(new DomToTreeModelAdapter());
	// Clear out the last selected DOM node pointer
	lastSelection = null;
    }

    public void setAnnotator(Annotator a) {
	if (a != null) {
	    annotatorLoggedIn = true;
	} else {
	    annotatorLoggedIn = false;
	}
	annPane.setAnnotator(a);
    }

    /** AdapterNode 
     *  Inner adapter class that takes a node from a DOM
     *  tree and returns a more readable string description of the
     *  node
     */
    public class AdapterNode 
    { 
	org.w3c.dom.Node domNode;
	
	// Construct an Adapter node from a DOM node
	public AdapterNode(org.w3c.dom.Node node) {
	    domNode = node;
	}

	// Return a string that identifies this node in the tree
	// *** Refer to table at top of org.w3c.dom.Node ***
	public String toString() {
	    String s = new String("");
	    /* If it is an element, we do not want to say Element: */
	    if (domNode.getNodeType() != ELEMENT_TYPE) {
		s = typeName[domNode.getNodeType()] + ": ";
	    }
	    String nodeName = domNode.getNodeName();
	    if (! nodeName.startsWith("#")) {
		s += nodeName;
		if (nodeName.equals("Chapter") ||
		    nodeName.equals("Page")) {
		    //  Append value of the attribute "Number" and :
		    org.w3c.dom.NamedNodeMap attrs = domNode.getAttributes();
		    s += " " + attrs.getNamedItem("Number").getNodeValue();
			
		    if (nodeName.equals("Chapter")) {
			s += ": ";
		    }			
		}		
	    }
	    if (props.getProperty("compressXMLTreeView", "true").equals("true")) {
		String t = content().trim();
		int x = t.indexOf("\n");
		if (x >= 0) t = t.substring(0, x);
		s += " " + t;
		return s;
	    }	    
	    if (domNode.getNodeValue() != null) {
		if (s.startsWith("ProcInstr")) 
		    s += ", "; 
		else 
		    s += ": ";
		// Trim the value to get rid of NL's at the front
		String t = domNode.getNodeValue().trim();
		int x = t.indexOf("\n");
		if (x >= 0) t = t.substring(0, x);
		s += t;
	    }
	    return s;
	}
	
	public String content() {
	    String s = new String("");
	    org.w3c.dom.NodeList nodeList = domNode.getChildNodes();
	    for (int i=0; i<nodeList.getLength(); i++) {
		org.w3c.dom.Node node = nodeList.item(i);
		int type = node.getNodeType();
		AdapterNode adpNode = new AdapterNode(node);
		if (type == ELEMENT_TYPE) {   
		    if (treeElement(node.getNodeName())) 
			continue;
		    /* We don't want to put the name of the elements up,
		       just the content */
		    s += adpNode.content();
		} else if (type == TEXT_TYPE) {
		    s += node.getNodeValue();
		} else if (type == ENTITYREF_TYPE) {
		    // The content is in the TEXT node under it
		    s += adpNode.content();
		} else if (type == CDATA_TYPE) {          
		    StringBuffer sb = new StringBuffer( node.getNodeValue() );
		    for (int j=0; j<sb.length(); j++) {
			if (sb.charAt(j) == '<') {
			    sb.setCharAt(j, '&');
			    sb.insert(j+1, "lt;");
			    j += 3;

			} else if (sb.charAt(j) == '>') {
			    sb.setCharAt(j, '&');
			    sb.insert(j+1, "gt;");
			    j += 3;
			} else if (sb.charAt(j) == '&') {
			    sb.setCharAt(j, '&');
			    sb.insert(j+1, "amp;");
			    j += 4;
			}
		    }
		    s += "<pre>" + sb + "\n</pre>";
		}
	    }
	    return s;
	}      
	public String nodeName() {
	    return domNode.getNodeName();
	}
	/**
	 * String getImageAttr
	 *  
	 * Return the value of the "Image" attribute, if it exists.
	 * return null otherwise.
	 */
	public String getImageAttr() {
	    String img;
	    org.w3c.dom.NamedNodeMap attrs = domNode.getAttributes();
	    if (attrs.getNamedItem("Image") != null) {
		img = attrs.getNamedItem("Image").getNodeValue();
	    } else {
		img = null;
	    }
	    return img;
	}
	public org.w3c.dom.Node getDOMNode() {
	    return domNode;
	}
	public int index(AdapterNode child) {
	    //System.err.println("Looking for index of " + child);
	    int count = childCount();
	    for (int i=0; i<count; i++) {
		AdapterNode n = this.child(i);
		if (child == n) return i;
	    }
	    return -1; // Should never get here.
	}
	
	public AdapterNode child(int searchIndex) {
	    //Note: JTree index is zero-based. 
	    org.w3c.dom.Node node = domNode.getChildNodes().item(searchIndex);
	    if (props.getProperty("compressXMLTreeView", "true").equals("true")) {
		// Return Nth displayable node
		int elementNodeIndex = 0;
		for (int i=0; i<domNode.getChildNodes().getLength(); i++) {
		    node = domNode.getChildNodes().item(i);
		    if (node.getNodeType() == ELEMENT_TYPE 
			&& treeElement( node.getNodeName() )
			&& elementNodeIndex++ == searchIndex) {
			break; 
		    }
		}
	    }	    
	    return new AdapterNode(node); 
	}
	
	public int childCount() {
	    if (props.getProperty("compressXMLTreeView", "true").equals("true")) {
		int count = 0;
		for (int i=0; i<domNode.getChildNodes().getLength(); i++) {
		    org.w3c.dom.Node node = domNode.getChildNodes().item(i); 
		    if (node.getNodeType() == ELEMENT_TYPE 
			&& treeElement( node.getNodeName() )) 
			{
			    ++count;
			}
		}
		return count;
	    } 
	    return domNode.getChildNodes().getLength();  
	}
	
    } // AdapterNode
    
    /**
     * DomToTreeModelAdapter
     *  This adapter converts the current Document (a DOM) into 
     *  a JTree model. 
     */
    public class DomToTreeModelAdapter implements javax.swing.tree.TreeModel {
	private Vector listenerList = new Vector(); // Vector of listeners
	// Basic TreeModel operations
	public Object getRoot() {
	    //System.err.println("Returning root: " +document);
	    if (document == null) {
		// If we have no document, create a new one
		/* Applet Security note: DocumentBuilderFactory.newInstance() tried to 
		   access the system property javax.xml.parsers.DocumentBuilderFactory.
		   What to do about this?  Can't test with appletViewer because of this. 
		*/
		DocumentBuilderFactory 
		    factory = DocumentBuilderFactory.newInstance();
		try {
		    DocumentBuilder builder = factory.newDocumentBuilder();
		    document = builder.newDocument();
		} catch (javax.xml.parsers.ParserConfigurationException pce) {
		    // Parser with specified options can't be built
		    pce.printStackTrace();
		}
	    }
	    return new AdapterNode(document);
	}
	public boolean isLeaf(Object aNode) {
	    // Determines whether the icon shows up to the left.
	    // Return true for any node with no children
	    AdapterNode node = (AdapterNode) aNode;
	    if (node.childCount() > 0) 
		return false;
	    return true;
	}
	public int getChildCount(Object parent) {
	    AdapterNode node = (AdapterNode) parent;
	    return node.childCount();
	}
	public Object getChild(Object parent, int index) {
	    AdapterNode node = (AdapterNode) parent;
	    return node.child(index);
	}
	public int getIndexOfChild(Object parent, Object child) {
	    AdapterNode node = (AdapterNode) parent;
	    return node.index((AdapterNode) child);
	}
	public void valueForPathChanged(TreePath path, Object newValue) {
	    // Null. We won't be making changes in the GUI
	    // If we did, we would ensure the new value was really new
	    // and then fire a TreeNodesChanged event.
	}
	public void addTreeModelListener(TreeModelListener listener) {
	    if (listener != null && !listenerList.contains(listener)) {
		listenerList.addElement(listener);
	    }
	}
	public void removeTreeModelListener(TreeModelListener listener) {
	    if (listener != null) {
		listenerList.removeElement(listener);
	    }
	}	
	public void fireTreeNodesChanged( TreeModelEvent e ) {
	    Enumeration listeners = listenerList.elements();
	    while ( listeners.hasMoreElements() ) {
		TreeModelListener listener = (TreeModelListener)listeners.nextElement();
		listener.treeNodesChanged( e );
	    }
	} 
	public void fireTreeNodesInserted( TreeModelEvent e ) {
	    Enumeration listeners = listenerList.elements();
	    while ( listeners.hasMoreElements() ) {
		TreeModelListener listener = (TreeModelListener)listeners.nextElement();
		listener.treeNodesInserted( e );
	    }
	}   
	public void fireTreeNodesRemoved( TreeModelEvent e ) {
	    Enumeration listeners = listenerList.elements();
	    while ( listeners.hasMoreElements() ) {
		TreeModelListener listener = (TreeModelListener)listeners.nextElement();
		listener.treeNodesRemoved( e );
	    }
	}   
	public void fireTreeStructureChanged( TreeModelEvent e ) {
	    Enumeration listeners = listenerList.elements();
	    while ( listeners.hasMoreElements() ) {
		TreeModelListener listener = (TreeModelListener)listeners.nextElement();
		listener.treeStructureChanged( e );
	    }
	}
    } // DomToTreeModelAdapter
    
} // XMLTreeDisplay

