
/**
 * GMAOOptionsFrame.java
 *
 *  This is a frame that has a tabbed-pane that lets us set up
 *  various GMAO options.  It will have a settings that more or less
 *  correspond to the settings in the GMAO.props file.
 *
 * Created: Sun Aug 06 14:54:30 2000
 *
 * @author David Evans
 * @version 0.2b
 */

/* 
   GMAO - Image Annotation for Manga Translation
   Copyright (C) 2000  David Kirk Evans
   
   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

   David Kirk Evans (devans@cs.columbia.edu)
   http://www.cs.columbia.edu/~devans/   
*/

import javax.swing.*;
import java.util.*;
import java.awt.*;
import java.awt.event.*;


public class GMAOOptionsFrame extends JFrame implements ActionListener {
    Properties props;
    DataOptionsPane dop;
    AppearanceOptionsPane app;
    GMAOLanguagePairs languageAnnotatorPairs;

    public GMAOOptionsFrame(Properties p) {
	props = p;
	getContentPane().setLayout(new BorderLayout());	
	// Create a tabbed pane for our center component
	JTabbedPane tpane = new JTabbedPane();
	dop = new DataOptionsPane();
	tpane.addTab("Files/Data", null, dop, "Set the options for where GMAO gets data from, and sends data to");
	app = new AppearanceOptionsPane();
	tpane.addTab("Appearance", null, app, "Set up GMAO Appearance options");
	getContentPane().add(tpane, BorderLayout.CENTER);
	// Create an "OK" and "Cancel" button
	Box buttonBox = new Box(BoxLayout.X_AXIS);
	JButton b = new JButton("OK");
	b.addActionListener(this);
	buttonBox.add(b);
	b = new JButton("Cancel");
	b.addActionListener(this);
	buttonBox.add(b);
	getContentPane().add(buttonBox, BorderLayout.SOUTH);
	// pack?
	pack();
	// Set the size and stuff?	
    }

    public void actionPerformed(ActionEvent e) {
	String command = e.getActionCommand();
	if (command.equals("OK")) {
	    // Make the changes 
	    dop.updateProps();
	    app.updateProps();
	    // Go away
	    hide();
	} else if (command.equals("Cancel")) {
	    // Just close the box
	    hide();
	}
    }
    
    public void setProperties(Properties p) {
	props = p;
    }

    public Properties getProperties() {
	return props;
    }

	/** 
	 *  setLanguagePairs(GMAOLanguagePairs lp) 
	 * 
	 *  Set the language pairs info thing
	 */
	public void setLanguagePairs(GMAOLanguagePairs lp) {
	    languageAnnotatorPairs = lp;
	    // Update the JComboBox in the appearance settings to take these into account
	    app.updateLanguagePairs();
	}
	
    /**
     *  Inner class that is a JPanel that sets up stuff for
     *  displaying the options associated with GMAO's data handling.
     */
    public class DataOptionsPane extends JPanel {
	JTextField imageHttpLocation;
	JTextField imageFileLocation;
	JLabel xmlFile;
	JTextField updateServer;
	String[] sourceArray = {"Local file system", "Web server"};
	JComboBox source;
	JTextField cacheSize;
	
	public DataOptionsPane() {
	    // Use two boxes for layout
	    Box lBox = new Box(BoxLayout.Y_AXIS);
	    Box fBox = new Box(BoxLayout.Y_AXIS);
	    // Add in component for imageSource
	    JLabel l = new JLabel("Image Source", SwingConstants.RIGHT);
	    l.setToolTipText("Select where images will be loaded from: the local file system, or a web server");
	    lBox.add(l);
	    source = new JComboBox(sourceArray);
	    if (props.getProperty("imageSource") != null && props.getProperty("imageSource").equals("file")) {
		source.setSelectedIndex(0);
	    } else {
		source.setSelectedIndex(1);
	    }
	    fBox.add(source);
	    l = new JLabel("Image web server URL:", SwingConstants.RIGHT);
	    l.setToolTipText("This is the URL for the web server that we will obtain images from.  Include the proper directories with trailing slash and protocal specifier (ie http://)");
	    lBox.add(l);
	    imageHttpLocation = new JTextField(props.getProperty("imageHttpLocation"));
	    fBox.add(imageHttpLocation);
	    l = new JLabel("Image file location:", SwingConstants.RIGHT);
	    l.setToolTipText("This is the directory that images should be loaded from if we are using the local file system.  Include trailing slash.");
	    lBox.add(l);
	    imageFileLocation = new JTextField(props.getProperty("imageFileLocation"));
	    fBox.add(imageFileLocation);
	    l = new JLabel("XML File:", SwingConstants.RIGHT);
	    l.setToolTipText("This is the URI for the XML file that we are using.  Change this using the \"Open\" menu.");
	    lBox.add(l);
	    xmlFile = new JLabel(props.getProperty("xmlFile"));
	    fBox.add(xmlFile);
	    l = new JLabel("Annotation Update server:", SwingConstants.RIGHT);
	    l.setToolTipText("This is the address (IP or fully qualified machine name) of the machine that we will send our annotations to.  It must be running a GMAOServer.  Leave this line blank if you do not want to send updates to a server.");
	    lBox.add(l);
	    updateServer = new JTextField(props.getProperty("updateServer") == null ? "" : props.getProperty("updateServer"));
	    fBox.add(updateServer);
	    l = new JLabel("Number of images to keep in cache:", SwingConstants.RIGHT);
	    l.setToolTipText("We will try to keep this many images in the GMAO image cache.  Note, setting this to a large value might lead to out-of-memory exceptions!");
	    lBox.add(l);
	    cacheSize = new JTextField(props.getProperty("cacheSize", "3"));
	    fBox.add(cacheSize);
	    // Add both boxes to us, in another box layout
	    Box mBox = new Box(BoxLayout.X_AXIS);
	    mBox.add(lBox);
	    mBox.add(fBox);
	    add(mBox);
	}

	/**
	 * updateProps() 
	 *  Look at the settings in this panel, and update
	 *  the properties file based on the settings.
	 */
	public void updateProps() {
	    if (((String)source.getSelectedItem()).equals("Local file system")) {
		props.setProperty("imageSource", "file");
	    } else {
		props.setProperty("imageSource", "http");
	    }
	    props.setProperty("imageHttpLocation", imageHttpLocation.getText());
	    props.setProperty("imageFileLocation", imageFileLocation.getText());
	    props.setProperty("updateServer", updateServer.getText());
	    try {
		int num = Integer.parseInt(cacheSize.getText());
		props.setProperty("cacheSize", ""+num);
	    } catch (NumberFormatException e) {
		// Caught it, just do nothing. 
	    }
	    // Done!
	}
    }

    /**
     * Inner class that is a JPanel that set up stuff for
     * displaying the options associated with GMAO's appearance.
     */
    public class AppearanceOptionsPane extends JPanel {
	ColorChoicePanel bgColor;
	ColorChoicePanel anFgColor;
	ColorChoicePanel anBgColor;
	JCheckBox drawAnnCkBox;
	JComboBox annLangBox;

	public AppearanceOptionsPane() {
	    // Use a vertical box for the labels
	    Box lBox = new Box(BoxLayout.Y_AXIS);
	    Box fBox = new Box(BoxLayout.Y_AXIS);
	    JLabel l = new JLabel("Empty background color:");
	    l.setToolTipText("This is the color that will be used as a background when we don't have an image in the annotation panel");
	    lBox.add(l);
	    bgColor = new ColorChoicePanel(new Color(Integer.parseInt(props.getProperty("Annotator.backgroundColor", "32767"))));
	    fBox.add(bgColor);
	    l = new JLabel("Annotation outline color:");
	    l.setToolTipText("The color of the annotation outlines, if we are drawing them");
	    lBox.add(l);
	    anFgColor = new ColorChoicePanel(new Color(Integer.parseInt(props.getProperty("Annotator.Annotation.fgcolor", "16000"))));
	    fBox.add(anFgColor);
	    l = new JLabel("Annotation background color:");
	    l.setToolTipText("The background color of the annotations.  None if you don't want to draw a background on the annotations");
	    lBox.add(l);
	    if ((props.getProperty("Annotator.Annotation.bgcolor") != null &&
		props.getProperty("Annotator.Annotation.bgcolor").equals("none")) || props.getProperty("Annotator.Annotation.bgcolor") == null) {
		anBgColor = new ColorChoicePanel(null);
	    } else {
		anBgColor = new ColorChoicePanel(new Color(Integer.parseInt(props.getProperty("Annotator.Annotation.bgcolor", "99999"))));
	    }
	    fBox.add(anBgColor);
	    l = new JLabel("Draw Annotation Outlines:");
	    l.setToolTipText("Draw outlines around annotations or not");
	    lBox.add(l);
	    if (props.getProperty("drawAnnotationOutlines") != null &&
		props.getProperty("drawAnnotationOutlines").equals("true")) {
		drawAnnCkBox = new JCheckBox("", true);
	    } else {
		drawAnnCkBox = new JCheckBox("", false);
	    }
	    fBox.add(drawAnnCkBox);
	    // Add in information about Annotator / language pairs to view
	    l = new JLabel("View Annotations by:");
	    l.setToolTipText("View annotations by this annotation in this language when available");
	    lBox.add(l);
	    String[] annLangArr = {"no annotators known yet - open up a file"};
	    if (languageAnnotatorPairs != null) {
		annLangArr = languageAnnotatorPairs.getAnnotatorLanguagePairs();
	    }
	    annLangBox = new JComboBox(annLangArr);
	    /* Set up the selection here by looking at properties for annotator and selection */
	    annLangBox.setSelectedIndex(getSelectedIndexForAnnLangPair(annLangArr));
	    fBox.add(annLangBox);
	    
	    // Join the label and field boxes together
	    Box mBox = new Box(BoxLayout.X_AXIS);
	    mBox.add(lBox);
	    mBox.add(fBox);
	    add(mBox);
	}

	/**
	 *  Using the properties, select the annotator language pair
	 *  that the user will be seeing.  The annotator language 
	 *  pairs are strings of the format ANNID (LANG) ###
	 */
	int getSelectedIndexForAnnLangPair(String[] arr) {
	    int num=0;
	    int index=0;
	    if (arr.length <= 1) {
		return 0;
	    }
	    String lang = props.getProperty("display.AnnotationLanguage", "English");
	    String annID = props.getProperty("display.Annotator", "");
	    // Search for the string with the right language and annotator.
	    while (num < arr.length) {
		int lastLParen = arr[num].lastIndexOf('(');
		int lastRParen = arr[num].lastIndexOf(')');
		if (lastLParen >= 0 && lastRParen >= 0) {
		    String arrAnn = arr[num].substring(0, lastLParen-1);
		    String arrLang = arr[num].substring(lastLParen+1, lastRParen);
		    if (lang.equals(arrLang)) {
			// We have the rigth language.  
			index = num;
			// Check to see if the annotator is right
			if (annID.equals(arrAnn)) {
			    // It matches.  Return this note.
			    return index;
			}
			// Not the right annotator.  Keep searching.  
		    }
		}
		// Check the next one
		num++;
	    }
	    return index;
	}

	/**
	 * updateProps() 
	 *  Look at the settings in this panel, and update
	 *  the properties file based on the settings.
	 */
	public void updateProps() {
	    if (bgColor.getColor() != null) {
		props.setProperty("Annotator.backgroundColor", ""+bgColor.getColor().getRGB());		
	    } 
	    if (anFgColor.getColor() != null) {
		props.setProperty("Annotator.Annotation.fgcolor", ""+anFgColor.getColor().getRGB());
	    }
	    if (anBgColor.getColor() != null) {
		props.setProperty("Annotator.Annotation.bgcolor", ""+anBgColor.getColor().getRGB());
	    } else {
		props.setProperty("Annotator.Annotation.bgcolor", "none");
	    }
	    if (drawAnnCkBox.isSelected()) {
		props.setProperty("drawAnnotationOutlines", "true");
	    } else {
		props.setProperty("drawAnnotationOutlines", "false");
	    } 
	    if (annLangBox.getSelectedIndex() != -1) {
		String selection = (String) annLangBox.getSelectedItem();
		// Have to parse out the selection.  
		// ANNID (LANGUAGE) NO CARE
		// Get the position of the last (
		int lastLParen = selection.lastIndexOf('(');
		int lastRParen = selection.lastIndexOf(')');
		String annID = selection.substring(0, lastLParen-1);
		String lang = selection.substring(lastLParen+1, lastRParen);
		// Set them in the properties.
		props.setProperty("display.AnnotationLanguage", lang);
		props.setProperty("display.Annotator", annID);
		// System.out.println("Set display.AnnotationLanguage "+lang);
		// System.out.println("Set display.Annotator "+annID);
	    } else {
		// It is -1.  Do nothing, let us use the defaults or
		// whatever we had before.
	    }
	    // Done!
	}
	
	/**
	 *  Change the model on the language pairs box to be current
	 */
	public void updateLanguagePairs() {
	    if (languageAnnotatorPairs != null) {
		annLangBox.setModel(new javax.swing.DefaultComboBoxModel(languageAnnotatorPairs.getAnnotatorLanguagePairs()));
	    }
	    // Set up the selection here too based on the properties!! 
	}
	
    }
    
    /**
     * Internal class to paint a color on itself, when 
     * clicked throw up a JColorChooser to pick a new color.
     */
    public class ColorChoicePanel extends JPanel implements MouseListener {
	Color c;  //  The color we are working with
	
	public ColorChoicePanel(Color color) {
	    c = color;
	    addMouseListener(this);
	}

	public void setColor(Color color) {
	    c = color;
	}

	public Color getColor() {
	    return c;
	}

	public void paintComponent(Graphics g) {
	    if (c != null) {
		g.setColor(c);
		Dimension ourSize = getSize();	    
		// Paint us in
		g.fillRect(0,0, ourSize.width, ourSize.height);
	    } else {
		g.setColor(Color.white);
		Dimension ourSize = getSize();	    
		// Paint us in
		g.fillRect(0,0, ourSize.width, ourSize.height);
		g.setColor(Color.black);
		g.drawString("None", 0, ourSize.height / 2);
	    }
	}
	
	public void mouseClicked(MouseEvent e) {
	    // Get a new color
	    c = JColorChooser.showDialog(this, "Please choose a new color", c);
	    repaint();
	}

	public void mouseEntered(MouseEvent e) {

	}

	public void mouseExited(MouseEvent e) {

	}

	public void mousePressed(MouseEvent e) {

	}
	
	public void mouseReleased(MouseEvent e) {

	}
    }
    
} // GMAOOptionsFrame




